class ItemNavigation:
    """
    Performs navigation through found results
    """

    def __init__(self, items):
        """
        :param list items: list of ResultItemWidget()'s
        """
        self.items = items
        self.items_num = len(items)
        self.selected = None

    def get_selected_index(self):
        return self.selected

    def select_default(self, query):
        """
        Selects item that should be selected by default
        If no such items found, select the first one in the list
        """
        indices = (index for index, item in enumerate(self.items) if item.selected_by_default(query))
        index = next(indices, 0)
        self.select(index)

    def select(self, index):
        if not 0 < index < self.items_num:
            index = 0

        if self.selected is not None:
            self.items[self.selected].deselect()

        self.selected = index
        self.items[index].select()

    def go_up(self):
        index = self.selected - 1 if self.selected is not None and self.selected > 0 else self.items_num - 1
        self.select(index)

    def go_down(self):
        index = self.selected + 1 if self.selected is not None and self.selected < self.items_num else 0
        self.select(index)

    def enter(self, query, index=None, alt=False):
        """
        Enter into selected item, unless 'index' is passed
        Return boolean - True if Ulauncher window should be closed
        """
        if index is not None:
            if not 0 <= index < self.items_num:
                raise IndexError

            self.select(index)
            return self.enter(query)

        if self.selected is not None:
            item = self.items[self.selected]
            action = item.on_enter(query) if not alt else item.on_alt_enter(query)
            if not action:
                return True
            action.run()
            return action.keep_app_open()

        return None



from threading import Thread
from functools import wraps


def run_async(*args, **kwargs):
    """
    Function decorator, intended to make "func" run in a new thread (asynchronously).

    :rtype: :class:`threading.Thread`

    Examples:

    >>> @run_async
    >>> def task1():
    >>>     do_something
    >>>
    >>> @run_async(daemon=True)
    >>> def task2():
    >>>     do_something_too
    >>>
    >>> t1 = task1()
    >>> t2 = task2()
    >>> ...
    >>> t1.join()
    >>> t2.join()

    """
    daemon = kwargs.get('daemon', False)

    def _run_async(func):
        @wraps(func)
        def async_func(*args, **kwargs):
            func_hl = Thread(target=func, args=args, kwargs=kwargs)
            func_hl.setDaemon(daemon)
            func_hl.start()
            return func_hl

        return async_func

    if len(args) == 1 and not kwargs and callable(args[0]):
        # No arguments, this is the decorator
        # Set default values for the arguments
        return _run_async(args[0])

    # This is just returning the decorator
    return _run_async



import os

from ulauncher.api.shared.action.OpenAction import OpenAction
from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.api.shared.action.SetUserQueryAction import SetUserQueryAction
from ulauncher.api.shared.item.SmallResultItem import SmallResultItem
from ulauncher.utils.Path import Path
from ulauncher.utils.image_loader import get_file_icon

from ulauncher.search.file_browser.FileQueries import FileQueries
from ulauncher.search.file_browser.alt_menu.CopyPathToClipboardItem import CopyPathToClipboardItem
from ulauncher.search.file_browser.alt_menu.OpenFolderItem import OpenFolderItem


class FileBrowserResultItem(SmallResultItem):
    """
    :param ~ulauncher.utils.Path.Path path:
    """

    # pylint: disable=super-init-not-called
    def __init__(self, path):
        self.path = path
        self._file_queries = FileQueries.get_instance()

    def get_name(self):
        """
        :return: name to show in the list
        """
        return self.path.get_basename()

    def get_name_highlighted(self, query, color):
        query = os.path.basename(query)
        return super().get_name_highlighted(query, color)

    def get_icon(self):
        return get_file_icon(self.path, self.get_icon_size())

    def on_enter(self, query):
        self._file_queries.save_query(self.path.get_abs_path())
        if self.path.is_dir():
            return SetUserQueryAction(os.path.join(self.path.get_user_path(), ''))

        return OpenAction(self.path.get_abs_path())

    def on_alt_enter(self, query):
        menu_items = self._get_dir_alt_menu() if self.path.is_dir() else self._get_file_alt_menu()
        return RenderResultListAction(menu_items)

    def _get_dir_alt_menu(self):
        """
        :rtype: list of ResultItems
        """
        open_folder = OpenFolderItem(self.path)
        open_folder.set_name("Open Folder '%s'" % self.path.get_basename())
        return [open_folder, CopyPathToClipboardItem(self.path)]

    def _get_file_alt_menu(self):
        """
        :rtype: list of ResultItems
        """
        open_folder = OpenFolderItem(Path(self.path.get_dirname()))
        open_folder.set_name('Open Containing Folder')
        return [open_folder, CopyPathToClipboardItem(self.path)]



import logging
from collections import defaultdict

from ulauncher.api.shared.Response import Response
from ulauncher.api.shared.action.BaseAction import BaseAction
from ulauncher.api.shared.event import PreferencesEvent, PreferencesUpdateEvent
from ulauncher.api.client.EventListener import EventListener
from ulauncher.api.client.Client import Client
from ulauncher.api.client.setup_logging import setup_logging


class Extension:
    """
    Manages extension runtime
    """

    def __init__(self):
        self._listeners = defaultdict(list)
        self._client = Client(self)
        self.preferences = {}
        self.logger = logging.getLogger(__name__)
        setup_logging()

    def subscribe(self, event_type, event_listener):
        """
        Example:

            extension.subscribe(PreferencesEvent, PreferencesEventListener())

        :param type event_type:
        :param ~ulauncher.api.client.EventListener.EventListener event_listener:
        """
        self._listeners[event_type].append(event_listener)

    def get_listeners_for_event(self, event):
        """
        :param ~ulauncher.api.shared.event.BaseEvent event:
        :rtype: ~ulauncher.api.client.EventListener.EventListener
        """
        return self._listeners[type(event)]

    def trigger_event(self, event):
        """
        :param ~ulauncher.api.shared.event.BaseEvent event:
        """
        listeners = self.get_listeners_for_event(event)
        if not listeners:
            self.logger.debug('No listeners for event %s', type(event).__name__)
            return

        for listener in listeners:
            action = listener.on_event(event, self)
            if action:
                assert isinstance(action, BaseAction), "on_event return value is not an instance of BaseAction"
                self._client.send(Response(event, action))

    def run(self):
        """
        Subscribes to events and connects to Ulauncher WS server
        """
        self.subscribe(PreferencesEvent, PreferencesEventListener())
        self.subscribe(PreferencesUpdateEvent, PreferencesUpdateEventListener())
        self._client.connect()


# pylint: disable=too-few-public-methods
class PreferencesEventListener(EventListener):

    def on_event(self, event, extension):
        extension.preferences.update(event.preferences)


class PreferencesUpdateEventListener(EventListener):

    def on_event(self, event, extension):
        extension.preferences[event.id] = event.new_value



import os
from typing import List, Union
import gi
gi.require_version('Gdk', '3.0')
# pylint: disable=wrong-import-position
from gi.repository import Gdk

from ulauncher.api.shared.action.BaseAction import BaseAction
from ulauncher.api.shared.action.DoNothingAction import DoNothingAction
from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.api.shared.action.SetUserQueryAction import SetUserQueryAction
from ulauncher.utils.Path import Path, InvalidPathError
from ulauncher.search.BaseSearchMode import BaseSearchMode
from ulauncher.search.SortedList import SortedList
from ulauncher.search.file_browser.FileBrowserResultItem import FileBrowserResultItem
from ulauncher.search.file_browser.FileQueries import FileQueries


class FileBrowserMode(BaseSearchMode):
    RESULT_LIMIT = 17

    def __init__(self):
        self._file_queries = FileQueries.get_instance()  # type: FileQueries

    def is_enabled(self, query: str) -> bool:
        """
        Enabled for queries like:
        ~/Downloads
        $USER/Downloads
        /usr/bin/foo
        """
        try:
            return query.lstrip()[0] in ('~', '/', '$')
        except IndexError:
            return False

    def list_files(self, path_str: str, sort_by_usage: bool = False) -> List[str]:
        files = os.listdir(path_str)

        def get_last_used_time(file: str) -> float:
            return self._file_queries.find(os.path.join(path_str, file)) or 0

        if sort_by_usage:
            return sorted(files, reverse=True, key=get_last_used_time)

        return sorted(files)

    def create_result_item(self, path_srt: str) -> FileBrowserResultItem:
        return FileBrowserResultItem(Path(path_srt))

    def filter_dot_files(self, file_list: List[str]) -> List[str]:
        return list(filter(lambda f: not f.startswith('.'), file_list))

    def handle_query(self, query: str) -> BaseAction:
        if query == '~':
            return SetUserQueryAction('~/')

        path = Path(query)
        result_items: Union[List, SortedList] = []

        try:
            existing_dir = path.get_existing_dir()

            if existing_dir == path.get_abs_path():
                file_names = self.list_files(path.get_abs_path(), sort_by_usage=True)
                for name in self.filter_dot_files(file_names)[:self.RESULT_LIMIT]:
                    file = os.path.join(existing_dir, name)
                    result_items.append(self.create_result_item(file))

            else:
                file_names = self.list_files(existing_dir)
                search_for = path.get_search_part()

                if not search_for.startswith('.'):
                    file_names = self.filter_dot_files(file_names)

                files = [os.path.join(existing_dir, name) for name in file_names]
                result_items = SortedList(search_for, min_score=40, limit=self.RESULT_LIMIT)
                result_items.extend([self.create_result_item(name) for name in reversed(files)])

        except (InvalidPathError, OSError):
            result_items = []

        return RenderResultListAction(result_items)

    def handle_key_press_event(self, widget, event, query):
        keyval = event.get_keyval()
        keyname = Gdk.keyval_name(keyval[1])
        ctrl = event.state & Gdk.ModifierType.CONTROL_MASK
        path = Path(query)
        if keyname == 'BackSpace' and not ctrl and widget.get_position() == len(query) and path.is_dir() and \
           not widget.get_selection_bounds():
            # stop key press event if:
            # it's a BackSpace key and
            # Ctrl modifier is not pressed and
            # cursor is at the last position and
            # path exists and it's a directory and
            # input text is not selected
            widget.emit_stop_by_name('key-press-event')
            return SetUserQueryAction(os.path.join(path.get_dirname(), ''))

        return DoNothingAction()



import webbrowser
from ulauncher.api.shared.action.BaseAction import BaseAction


class OpenUrlAction(BaseAction):
    """
    Opens URL in a default browser

    :param str url:
    """

    def __init__(self, url):
        self.url = url

    def run(self):
        webbrowser.open_new_tab(self.url)



import gi
gi.require_version('Gtk', '3.0')
# pylint: disable=wrong-import-position
from gi.repository import Gtk


def gtk_version_is_gte(major: int, minor: int, micro: int) -> bool:
    gtk_version = (Gtk.get_major_version(), Gtk.get_minor_version(), Gtk.get_micro_version())
    return gtk_version >= (major, minor, micro)



import os
from uuid import uuid4
from time import time
from ulauncher.config import CONFIG_DIR, get_default_shortcuts
from ulauncher.utils.db.KeyValueJsonDb import KeyValueJsonDb
from ulauncher.utils.decorator.singleton import singleton


class ShortcutsDb(KeyValueJsonDb):

    @classmethod
    @singleton
    def get_instance(cls):
        db_path = os.path.join(CONFIG_DIR, 'shortcuts.json')
        is_first_run = not os.path.exists(db_path)
        db = cls(db_path)
        db.open()

        if is_first_run:
            db.set_records(get_default_shortcuts())

        db.commit()

        return db

    def commit(self):
        self.ensure_user_paths()
        super().commit()

    def get_sorted_records(self):
        return [rec for rec in sorted(self.get_shortcuts(), key=lambda rec: rec['added'])]

    def get_shortcuts(self):
        for rec in self.get_records().values():
            # fix for shortcuts generated in v6 branch not working in v5
            if rec.get("icon") and not os.path.isfile(rec.get("icon")):
                legacy_path = (
                    rec.get("icon", "")
                    .replace("/icons/", "/media/")
                    .replace(".svg", "-icon.svg")
                    .replace(".png", "-icon.png")
                )
                if os.path.isfile(legacy_path):
                    rec["icon"] = legacy_path
            yield rec

    # pylint: disable=too-many-arguments
    def put_shortcut(self, name, keyword, cmd, icon, is_default_search, run_without_argument, id=None):
        """
        If id is not provided it will be generated using uuid4() function
        """
        id = id or str(uuid4())
        self._records[id] = {
            "id": id,
            "name": name,
            "keyword": keyword,
            "cmd": cmd,
            "icon": icon,
            "is_default_search": bool(is_default_search),
            "run_without_argument": bool(run_without_argument),
            # use previously added time if record with the same id exists
            "added": self._records.get(id, {"added": time()})["added"]
        }
        return id

    def ensure_user_paths(self):
        for s in self.get_shortcuts():
            s['icon'] = get_user_path(s.get('icon'))


def get_user_path(path):
    user_home = os.path.expanduser('~')
    if path and path.startswith(user_home):
        return path.replace(user_home, '~', 1)

    return path



import pickle


class BaseEvent:

    def __eq__(self, other):
        return pickle.dumps(self) == pickle.dumps(other)

    def __ne__(self, other):
        return pickle.dumps(self) != pickle.dumps(other)


class KeywordQueryEvent(BaseEvent):
    """
    Is triggered when user enters query that starts with your keyword + Space

    :param ~ulauncher.search.Query.Query query:
    """

    def __init__(self, query):
        self.query = query

    def get_keyword(self):
        """
        :rtype: str
        """
        return self.query.get_keyword()

    def get_query(self):
        """
        :rtype: :class:`~ulauncher.search.Query.Query`
        """
        return self.query

    def get_argument(self):
        """
        :rtype: str
        :returns: None if arguments were not specified
        """
        return self.query.get_argument()


# pylint: disable=too-few-public-methods
class ItemEnterEvent(BaseEvent):
    """
    Is triggered when selected item has action of type :class:`~ulauncher.api.shared.action.ExtensionCustomAction`
    Whatever data you've passed to action will be available in in this class using method :meth:`get_data`

    :param str data:
    """

    def __init__(self, data):
        self._data = data

    def get_data(self):
        """
        :returns: whatever object you have passed to :class:`~ulauncher.api.shared.action.ExtensionCustomAction`
        """
        return pickle.loads(self._data)


class SystemExitEvent(BaseEvent):
    """
    Is triggered when extension is about to be terminated.

    Your extension has 300ms to handle this event and shut down properly.
    After that it will be terminated with SIGKILL
    """


class PreferencesUpdateEvent(BaseEvent):
    """
    Is triggered when user updates preference through Preferences window

    :param str id:
    :param str old_value:
    :param str new_value:
    """

    id = None
    old_value = None
    new_value = None

    def __init__(self, id, old_value, new_value):
        self.id = id
        self.old_value = old_value
        self.new_value = new_value


class PreferencesEvent(BaseEvent):
    """
    Is triggered on start

    :param dict preferences:
    """

    preferences = None

    def __init__(self, preferences):
        self.preferences = preferences



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: Bind a new key
import os
import logging
import json
import mimetypes
from urllib.parse import unquote
from typing import List, Optional, cast
import time
import traceback

import gi
gi.require_version('Gio', '2.0')
gi.require_version('GLib', '2.0')
gi.require_version('Gtk', '3.0')

# pylint: disable=wrong-import-position,unused-argument
from gi.repository import Gio, Gtk, GLib  # type: ignore

from ulauncher.api.shared.action.OpenAction import OpenAction
from ulauncher.ui.windows.HotkeyDialog import HotkeyDialog
from ulauncher.ui.windows.WindowHelper import WindowHelper
from ulauncher.ui.windows.Builder import Builder
from ulauncher.api.version import api_version
from ulauncher.api.shared.event import PreferencesUpdateEvent
from ulauncher.api.server.extension_finder import find_extensions
from ulauncher.api.server.ExtensionPreferences import ExtensionPreferences, PreferenceItems
from ulauncher.api.server.ExtensionDb import ExtensionDb
from ulauncher.api.server.ExtensionRunner import ExtensionRunner, ExtRunError
from ulauncher.api.server.ExtensionManifest import ExtensionManifestError
from ulauncher.api.server.ExtensionDownloader import (ExtensionDownloader, ExtensionIsUpToDateError)
from ulauncher.api.shared.errors import UlauncherAPIError, ErrorName
from ulauncher.api.server.ExtensionServer import ExtensionServer
from ulauncher.utils.Theme import themes, Theme, load_available_themes
from ulauncher.utils.decorator.glib_idle_add import glib_idle_add
from ulauncher.utils.mypy_extensions import TypedDict
from ulauncher.utils.decorator.run_async import run_async
from ulauncher.utils.wayland import is_wayland
from ulauncher.utils.Settings import Settings
from ulauncher.utils.Router import Router, get_url_params
from ulauncher.utils.AutostartPreference import AutostartPreference
from ulauncher.utils.WebKit2 import WebKit2
from ulauncher.ui.AppIndicator import AppIndicator
from ulauncher.search.shortcuts.ShortcutsDb import ShortcutsDb
from ulauncher.config import get_data_file, get_options, get_version, EXTENSIONS_DIR


logger = logging.getLogger(__name__)
rt = Router()


class PrefsApiError(UlauncherAPIError):
    pass


ExtError = TypedDict('ExtError', {
    'errorName': str,
    'message': str
})

ExtensionInfo = TypedDict('ExtensionInfo', {
    'id': str,
    'url': str,
    'updated_at': str,
    'last_commit': str,
    'last_commit_time': str,
    'name': str,
    'icon': str,
    'description': str,
    'developer_name': str,
    'is_running': bool,
    'runtime_error': Optional[ExtRunError],
    'preferences': PreferenceItems,
    'error': Optional[ExtError]
})


# pylint: disable=too-many-instance-attributes, too-many-public-methods, attribute-defined-outside-init
class PreferencesUlauncherDialog(Gtk.Dialog, WindowHelper):
    __gtype_name__ = "PreferencesUlauncherDialog"

    _hotkey_name = None

    ######################################
    # Initialization
    ######################################

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated PreferencesDialog object.
        """
        builder = Builder.new_from_file('PreferencesUlauncherDialog')
        new_object = builder.get_object("preferences_ulauncher_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initializing should be called after parsing the ui definition
        and creating a PreferencesDialog object with it in order to
        finish initializing the start of the new PreferencesUlauncherDialog
        instance.

        Put your initialization code in here and leave __init__ undefined.
        """

        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self, True)

        # unnecessary action area can be removed only manually, like this
        self.ui['dialog_action_area'].destroy()

        self.settings = Settings.get_instance()
        self._init_webview()
        self.init_styles(get_data_file('styles', 'preferences.css'))
        self._handle_no_window_shadow_option(self.ui['window_wrapper'])
        self.autostart_pref = AutostartPreference()
        self.hotkey_dialog = HotkeyDialog()
        self.hotkey_dialog.connect('hotkey-set', self.on_hotkey_set)

        self.show_all()

    def _handle_no_window_shadow_option(self, window_wrapper):
        # removing window shadow solves issue with DEs without a compositor (#230)
        if get_options().no_window_shadow:
            window_wrapper.get_style_context().add_class('no-window-shadow')

    def _init_webview(self):
        """
        Initialize preferences WebView
        """
        settings = WebKit2.Settings(
            enable_developer_extras=bool(get_options().dev),
            enable_hyperlink_auditing=False,
            enable_page_cache=False,
            enable_webgl=False,
            enable_write_console_messages_to_stdout=True,
            enable_xss_auditor=False,
            hardware_acceleration_policy=WebKit2.HardwareAccelerationPolicy.NEVER,
        )
        self.webview = WebKit2.WebView(settings=settings)
        self.ui['scrolled_window'].add(self.webview)
        self._load_prefs_html()

        self.webview.get_context().register_uri_scheme('prefs', self.on_scheme_callback)
        self.webview.get_context().register_uri_scheme('file2', self.serve_file)
        self.webview.get_context().set_cache_model(WebKit2.CacheModel.DOCUMENT_VIEWER)  # disable caching
        self.webview.connect('button-press-event', self.webview_on_button_press_event)
        self.webview.connect('context-menu', self.webview_on_context_menu)

    ######################################
    # Overrides
    ######################################

    # pylint: disable=arguments-differ
    def present(self, page):
        self._load_prefs_html(page)
        super().present()

    def show(self, page):
        self._load_prefs_html(page)
        super().show()

    ######################################
    # GTK event handlers
    ######################################
    #
    def webview_on_button_press_event(self, widget, event):
        """
        Makes preferences window draggable by empty an empty space between navigation and close button
        also by the color stripe
        """
        window_width = self.get_size()[0]
        if event.button == 1 and (690 < event.x < window_width - 100 and 0 < event.y < 69) or event.y <= 11:
            self.begin_move_drag(event.button, event.x_root, event.y_root, event.time)

        return False

    def webview_on_context_menu(self, *args):
        return bool(not get_options().dev)

    ######################################
    # WebView communication methods
    ######################################

    @run_async
    def serve_file(self, scheme_request):
        """
        Serves file with custom file2:// protocol because file:// breaks for some
        """
        # pylint: disable=broad-except
        try:
            params = get_url_params(scheme_request.get_uri())
            file_path = params['path'].replace("null/", "/").split("#", 1)[0]

            mime_type = mimetypes.guess_type(file_path)[0]
            stream = Gio.file_new_for_path(file_path).read()
            scheme_request.finish(stream, -1, mime_type)
        except Exception as e:
            logger.exception('Unable to send file. %s: %s', type(e).__name__, e)
            return

    @run_async
    def on_scheme_callback(self, scheme_request):
        """
        Handles Javascript-to-Python calls
        """

        # pylint: disable=broad-except
        try:
            params = get_url_params(scheme_request.get_uri())
            callback_name = params['query']['callback']
            assert callback_name
        except Exception as e:
            logger.exception('API call failed. %s: %s', type(e).__name__, e)
            return

        try:
            resp = rt.dispatch(self, scheme_request.get_uri())
            callback = '%s(%s);' % (callback_name, json.dumps(resp))
        except UlauncherAPIError as e:
            error_type = type(e).__name__
            logger.error('%s: %s', error_type, e)
            callback = '%s(null, %s);' % (callback_name, json.dumps({
                'message': str(e),
                'type': error_type,
                'errorName': e.error_name,
                'stacktrace': traceback.format_exc()
            }))
        except Exception as e:
            message = 'Unexpected API error. %s: %s' % (type(e).__name__, e)
            logger.exception(message)
            callback = '%s(null, %s);' % (callback_name, json.dumps({
                'message': str(e),
                'type': type(e).__name__,
                'errorName': ErrorName.UnhandledError.value,
                'stacktrace': traceback.format_exc()
            }))

        try:
            stream = Gio.MemoryInputStream.new_from_data(callback.encode())
            # send response
            scheme_request.finish(stream, -1, 'text/javascript')
        except Exception as e:
            logger.exception('Unexpected API error. %s: %s', type(e).__name__, e)

    def send_webview_notification(self, name, data):
        self.webview.run_javascript('onNotification("%s", %s)' % (name, json.dumps(data)))

    ######################################
    # Request handlers
    ######################################

    @rt.route('/get/all')
    def prefs_get_all(self, url_params):
        logger.info('API call /get/all')
        return {
            'show_indicator_icon': self.settings.get_property('show-indicator-icon'),
            'hotkey_show_app': self.get_app_hotkey(),
            'autostart_allowed': self.autostart_pref.is_allowed(),
            'autostart_enabled': self.autostart_pref.is_on(),
            'show_recent_apps': self.settings.get_property('show-recent-apps'),
            'clear_previous_query': self.settings.get_property('clear-previous-query'),
            'blacklisted_desktop_dirs': self.settings.get_property('blacklisted-desktop-dirs'),
            'disable_desktop_filters': self.settings.get_property('disable-desktop-filters'),
            'available_themes': self._get_available_themes(),
            'theme_name': Theme.get_current().get('name'),
            'render_on_screen': self.settings.get_property('render-on-screen'),
            'is_wayland': is_wayland(),
            'terminal_command': self.settings.get_property('terminal-command'),
            'grab_mouse_pointer': self.settings.get_property('grab-mouse-pointer'),
            'env': {
                'version': get_version(),
                'api_version': api_version,
                'user_home': os.path.expanduser('~')
            }
        }

    @rt.route('/set/show-indicator-icon')
    def prefs_set_show_indicator_icon(self, url_params):
        show_indicator = self._get_bool(url_params['query']['value'])
        logger.info('Set show-indicator-icon to %s', show_indicator)
        self.settings.set_property('show-indicator-icon', show_indicator)
        self.settings.save_to_file()
        indicator = AppIndicator.get_instance()
        GLib.idle_add(indicator.switch, show_indicator)

    @rt.route('/set/autostart-enabled')
    def prefs_set_autostart(self, url_params):
        is_on = self._get_bool(url_params['query']['value'])
        logger.info('Set autostart-enabled to %s', is_on)
        if is_on and not self.autostart_pref.is_allowed():
            raise PrefsApiError("Unable to turn on autostart preference")

        try:
            self.autostart_pref.switch(is_on)
        except Exception as e:
            raise PrefsApiError('Caught an error while switching "autostart": %s' % e) from e

    @rt.route('/set/show-recent-apps')
    def prefs_set_show_recent_apps(self, url_params):
        try:
            recent_apps_number = int(url_params['query']['value'])
        except ValueError:
            recent_apps_number = 3
        logger.info('Set show-recent-apps to %s', recent_apps_number)
        self.settings.set_property('show-recent-apps', recent_apps_number)
        self.settings.save_to_file()

    @rt.route('/set/hotkey-show-app')
    @glib_idle_add
    def prefs_set_hotkey_show_app(self, url_params):
        hotkey = url_params['query']['value']
        logger.info('Set hotkey-show-app to %s', hotkey)

        # Bind a new key
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow
        ulauncher_window = UlauncherWindow.get_instance()
        ulauncher_window.bind_show_app_hotkey(hotkey)
        self.settings.set_property('hotkey-show-app', hotkey)
        self.settings.save_to_file()

    @rt.route('/set/theme-name')
    @glib_idle_add
    def prefs_set_theme_name(self, url_params):
        name = url_params['query']['value']
        logger.info('Set theme-name to %s', name)

        self.settings.set_property('theme-name', name)
        self.settings.save_to_file()

        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow
        ulauncher_window = UlauncherWindow.get_instance()
        ulauncher_window.init_theme()

    @rt.route('/set/terminal-command')
    def prefs_set_terminal_command(self, url_params):
        terminal_command = url_params['query']['value']
        logger.info('Set terminal launch command to %s', terminal_command)
        self.settings.set_property('terminal-command', terminal_command)
        self.settings.save_to_file()

    @rt.route('/show/hotkey-dialog')
    @glib_idle_add
    def prefs_showhotkey_dialog(self, url_params):
        self._hotkey_name = url_params['query']['name']
        logger.info('Show hotkey-dialog for %s', self._hotkey_name)
        self.hotkey_dialog.present()

    @rt.route('/set/clear-previous-query')
    def prefs_set_clear_previous_text(self, url_params):
        is_on = self._get_bool(url_params['query']['value'])
        logger.info('Set clear-previous-query to %s', is_on)
        self.settings.set_property('clear-previous-query', is_on)
        self.settings.save_to_file()

    @rt.route('/set/grab-mouse-pointer')
    def prefs_set_grab_mouse_pointer(self, url_params):
        is_on = self._get_bool(url_params['query']['value'])
        logger.info('Set grab-mouse-pointer to %s', is_on)
        self.settings.set_property('grab-mouse-pointer', is_on)
        self.settings.save_to_file()

    @rt.route('/set/blacklisted-desktop-dirs')
    def prefs_set_blacklisted_desktop_dirs(self, url_params):
        dirs = url_params['query']['value']
        logger.info('Set blacklisted-desktop-dirs to %s', dirs)
        self.settings.set_property('blacklisted-desktop-dirs', dirs)
        self.settings.save_to_file()

    @rt.route('/set/disable-desktop-filters')
    def prefs_set_disable_desktop_filters(self, url_params):
        is_on = self._get_bool(url_params['query']['value'])
        logger.info('Set disable-desktop-filters to %s', is_on)
        self.settings.set_property('disable-desktop-filters', is_on)
        self.settings.save_to_file()

    @rt.route('/set/render-on-screen')
    def prefs_set_render_on_screen(self, url_params):
        selected_option = url_params['query']['value']
        logger.info('Set render-on-screen to %s', selected_option)
        self.settings.set_property('render-on-screen', selected_option)
        self.settings.save_to_file()

    @rt.route('/show/file-browser')
    @glib_idle_add
    def prefs_show_file_browser(self, url_params):
        """
        Request params: type=(image|all), name=(str)
        """
        file_browser_name = url_params['query']['name']
        logger.info('Show file browser dialog for %s', file_browser_name)
        dialog = Gtk.FileChooserDialog("Please choose a file", self, Gtk.FileChooserAction.OPEN,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        filter_images = Gtk.FileFilter()
        filter_images.set_name("Image files")
        filter_images.add_mime_type("image/*")
        dialog.add_filter(filter_images)

        response = dialog.run()
        data = {
            'value': None
        }
        if response == Gtk.ResponseType.OK:
            data['value'] = dialog.get_filename()

        logger.debug('%s %s', file_browser_name, data)
        self.send_webview_notification(file_browser_name, data)
        dialog.destroy()

    @rt.route('/open/web-url')
    def prefs_open_url(self, url_params):
        url = unquote(url_params['query']['url'])
        logger.info('Open Web URL %s', url)
        OpenAction(url).run()

    @rt.route('/open/extensions-dir')
    def open_extensions_dir(self, _):
        logger.info('Open extensions directory "%s" in default file manager.', EXTENSIONS_DIR)
        OpenAction(EXTENSIONS_DIR).run()

    @rt.route('/close')
    def prefs_close(self, url_params):
        logger.info('Close preferences')
        self.hide()

    @rt.route('/shortcut/get-all')
    def prefs_shortcut_get_all(self, url_params):
        logger.info('Handling /shortcut/get-all')
        shortcuts = ShortcutsDb.get_instance()
        return shortcuts.get_sorted_records()

    @rt.route('/shortcut/update')
    @rt.route('/shortcut/add')
    def prefs_shortcut_update(self, url_params):
        req_data = url_params['query']
        logger.info('Add/Update shortcut: %s', json.dumps(req_data))
        shortcuts = ShortcutsDb.get_instance()
        id = shortcuts.put_shortcut(req_data['name'],
                                    req_data['keyword'],
                                    req_data['cmd'],
                                    req_data.get('icon') or None,
                                    str_to_bool(req_data['is_default_search']),
                                    str_to_bool(req_data['run_without_argument']),
                                    req_data.get('id'))
        shortcuts.commit()
        return {'id': id}

    @rt.route('/shortcut/remove')
    def prefs_shortcut_remove(self, url_params):
        req_data = url_params['query']
        logger.info('Remove shortcut: %s', json.dumps(req_data))
        shortcuts = ShortcutsDb.get_instance()
        shortcuts.remove(req_data['id'])
        shortcuts.commit()

    @rt.route('/extension/get-all')
    def prefs_extension_get_all(self, url_params):
        logger.info('Handling /extension/get-all')
        return self._get_all_extensions()

    @rt.route('/extension/add')
    def prefs_extension_add(self, url_params):
        url = url_params['query']['url']
        logger.info('Add extension: %s', url)
        downloader = ExtensionDownloader.get_instance()
        ext_id = downloader.download(url)
        ExtensionRunner.get_instance().run(ext_id)

        # Looping until either runner.is_running() or runner.get_extension_error() returns something would be better
        # to avoid race condition and needless waiting
        time.sleep(1)

        return self._get_all_extensions()

    @rt.route('/extension/update-prefs')
    def prefs_extension_update_prefs(self, url_params):
        query = url_params['query']
        ext_id = query['id']
        logger.info('Update extension preferences: %s', query)
        prefix = 'pref.'
        controller = ExtensionServer.get_instance().get_controller(ext_id)
        if not controller:
            logger.error("Cannot update preferences. The extension controller is not registered. \
            Try relaunching or restarting Ulauncher, and check the extension readme for external dependencies \
            you may have to install manually")
            return
        preferences = [(key[len(prefix):], value) for key, value in query.items() if key.startswith(prefix)]
        for pref_id, value in preferences:
            old_value = controller.preferences.get(pref_id)['value']
            controller.preferences.set(pref_id, value)
            if value != old_value:
                controller.trigger_event(PreferencesUpdateEvent(pref_id, old_value, value))

    @rt.route('/extension/check-updates')
    def prefs_extension_check_updates(self, url_params):
        logger.info('Handling /extension/check-updates')
        ext_id = url_params['query']['id']
        try:
            return ExtensionDownloader.get_instance().get_new_version(ext_id)
        except ExtensionIsUpToDateError:
            return None

    @rt.route('/extension/update-ext')
    def prefs_extension_update_ext(self, url_params):
        ext_id = url_params['query']['id']
        logger.info('Update extension: %s', ext_id)
        downloader = ExtensionDownloader.get_instance()
        try:
            downloader.update(ext_id)
        except ExtensionManifestError as e:
            raise PrefsApiError(e)

    @rt.route('/extension/remove')
    def prefs_extension_remove(self, url_params):
        ext_id = url_params['query']['id']
        logger.info('Remove extension: %s', ext_id)
        downloader = ExtensionDownloader.get_instance()
        downloader.remove(ext_id)

    ######################################
    # Helpers
    ######################################

    def _get_all_extensions(self) -> List[ExtensionInfo]:
        extensions = []
        for ext_id, _ in find_extensions(EXTENSIONS_DIR):
            prefs = ExtensionPreferences.create_instance(ext_id)  # type: ExtensionPreferences
            prefs.manifest.refresh()
            error = None
            try:
                prefs.manifest.validate()
                prefs.manifest.check_compatibility()
            except UlauncherAPIError as e:
                error = cast(ExtError, {
                    'message': str(e),
                    'errorName': e.error_name
                })
            except Exception as e:
                error = cast(ExtError, {
                    'message': str(e),
                    'errorName': ErrorName.UnexpectedError.value
                })
            extensions.append(self._get_extension_info(ext_id, prefs, error))

        return extensions

    def _get_extension_info(self, ext_id: str, prefs: ExtensionPreferences, error: ExtError = None) -> ExtensionInfo:
        ext_db = ExtensionDb.get_instance()
        is_connected = bool(ExtensionServer.get_instance().get_controller(ext_id))
        ext_runner = ExtensionRunner.get_instance()
        is_running = is_connected or ext_runner.is_running(ext_id)
        ext_db_record = ext_db.find(ext_id, {})
        return {
            'id': ext_id,
            'url': ext_db_record.get('url'),
            'updated_at': ext_db_record.get('updated_at'),
            'last_commit': ext_db_record.get('last_commit'),
            'last_commit_time': ext_db_record.get('last_commit_time'),
            'name': prefs.manifest.get_name(),
            'icon': prefs.manifest.get_icon_path(),
            'description': prefs.manifest.get_description(),
            'developer_name': prefs.manifest.get_developer_name(),
            'preferences': prefs.get_items(),
            'error': error,
            'is_running': is_running,
            'runtime_error': ext_runner.get_extension_error(ext_id) if not is_running else None
        }

    def _load_prefs_html(self, page=''):
        uri = "file2://%s#/%s" % (get_data_file('preferences/dist/index.html'), page)
        self.webview.load_uri(uri)

    def _get_bool(self, str_val):
        return str(str_val).lower() in ('true', '1', 'on')

    def _get_available_themes(self):
        load_available_themes()
        return [dict(value=th.get('name'), text=th.get('display_name')) for th in themes.values()]

    def get_app_hotkey(self):
        app_hotkey_current_accel_name = self.settings.get_property('hotkey-show-app')
        try:
            (key, mode) = Gtk.accelerator_parse(app_hotkey_current_accel_name)
        # pylint: disable=broad-except
        except Exception:
            logger.warning('Unable to parse accelerator "%s". Use Ctrl+Space', app_hotkey_current_accel_name)
            (key, mode) = Gtk.accelerator_parse("<Primary>space")
        return Gtk.accelerator_get_label(key, mode)

    def on_hotkey_set(self, widget, hotkey_val, hotkey_display_val):
        self.send_webview_notification(self._hotkey_name, {
            'value': hotkey_val,
            'displayValue': hotkey_display_val
        })


def str_to_bool(value):
    return value in [1, 'true', 'True']



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
import os
import time
import logging
import threading

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
gi.require_version('GLib', '2.0')
gi.require_version('Keybinder', '3.0')

# pylint: disable=wrong-import-position, unused-argument
from gi.repository import Gtk, Gdk, GLib, Keybinder  # type: ignore

# pylint: disable=unused-import
# these imports are needed for Gtk to find widget classes
from ulauncher.ui.ResultItemWidget import ResultItemWidget  # noqa: F401
from ulauncher.ui.SmallResultItemWidget import SmallResultItemWidget   # noqa: F401

from ulauncher.config import get_data_file, get_options
from ulauncher.ui.ItemNavigation import ItemNavigation
from ulauncher.search.Search import Search
from ulauncher.search.apps.AppStatDb import AppStatDb
from ulauncher.api.server.ExtensionRunner import ExtensionRunner
from ulauncher.api.server.ExtensionServer import ExtensionServer
from ulauncher.api.server.ExtensionDownloader import ExtensionDownloader
from ulauncher.utils.AppCacheDb import AppCacheDb
from ulauncher.utils.Settings import Settings
from ulauncher.utils.decorator.singleton import singleton
from ulauncher.utils.display import get_current_screen_geometry, get_primary_screen_geometry, get_monitor_scale_factor
from ulauncher.utils.image_loader import load_image
from ulauncher.utils.version_cmp import gtk_version_is_gte
from ulauncher.utils.desktop.notification import show_notification
from ulauncher.utils.wayland import is_wayland_compatibility_on
from ulauncher.utils.Theme import Theme, load_available_themes
from ulauncher.search.apps.app_watcher import start as start_app_watcher
from ulauncher.search.Query import Query
from ulauncher.ui.windows.Builder import Builder
from ulauncher.ui.windows.WindowHelper import WindowHelper
from ulauncher.ui.windows.PreferencesUlauncherDialog import PreferencesUlauncherDialog

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes, too-many-public-methods, attribute-defined-outside-init
class UlauncherWindow(Gtk.Window, WindowHelper):
    __gtype_name__ = "UlauncherWindow"

    _current_accel_name = None
    _results_render_time = 0

    @classmethod
    @singleton
    def get_instance(cls):
        return cls()

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated BaseUlauncherWindow object.
        """
        builder = Builder.new_from_file('UlauncherWindow')
        new_object = builder.get_object("ulauncher_window")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initializing should be called after parsing the UI definition
        and creating a UlauncherWindow object with it in order to finish
        initializing the start of the new UlauncherWindow instance.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self, True)
        self.PreferencesDialog = None  # class
        self.preferences_dialog = None  # instance

        self.results_nav = None
        self.window = self.ui['ulauncher_window']
        self.window_body = self.ui['body']
        self.input = self.ui['input']
        self.prefs_btn = self.ui['prefs_btn']
        self.result_box = self.ui["result_box"]

        self.input.connect('changed', self.on_input_changed)
        self.prefs_btn.connect('clicked', self.on_mnu_preferences_activate)

        self.set_keep_above(True)

        self.PreferencesDialog = PreferencesUlauncherDialog
        self.settings = Settings.get_instance()

        self.fix_window_width()
        self.position_window()
        self.init_theme()

        # this will trigger to show frequent apps if necessary
        self.show_results([])

        if not is_wayland_compatibility_on():
            # bind hotkey
            Keybinder.init()
            accel_name = self.settings.get_property('hotkey-show-app')
            # bind in the main thread
            GLib.idle_add(self.bind_show_app_hotkey, accel_name)

        start_app_watcher()
        ExtensionServer.get_instance().start()
        time.sleep(0.01)
        ExtensionRunner.get_instance().run_all()
        if not get_options().no_extensions:
            ExtensionDownloader.get_instance().download_missing()

    ######################################
    # GTK Signal Handlers
    ######################################

    # pylint: disable=unused-argument
    def on_mnu_about_activate(self, widget, data=None):
        """Display the about page for ulauncher."""
        self.activate_preferences(page='about')

    def on_mnu_preferences_activate(self, widget, data=None):
        """Display the preferences window for ulauncher."""
        self.activate_preferences(page='preferences')

    def on_mnu_close_activate(self, widget, data=None):
        """Signal handler for closing the UlauncherWindow."""
        self.destroy()

    def on_destroy(self, widget, data=None):
        """Called when the UlauncherWindow is closed."""
        # Clean up code for saving application state should be added here.
        Gtk.main_quit()

    def on_preferences_dialog_destroyed(self, widget, data=None):
        '''only affects GUI

        logically there is no difference between the user closing,
        minimizing or ignoring the preferences dialog'''
        logger.debug('on_preferences_dialog_destroyed')
        # to determine whether to create or present preferences_dialog
        self.preferences_dialog = None

    def on_focus_out_event(self, widget, event):
        # apparently Gtk doesn't provide a mechanism to tell if window is in focus
        # this is a simple workaround to avoid hiding window
        # when user hits Alt+key combination or changes input source, etc.
        self.is_focused = False
        t = threading.Timer(0.07, lambda: self.is_focused or self.hide())
        t.start()

    def on_focus_in_event(self, *args):
        if self.settings.get_property('grab-mouse-pointer'):
            ptr_dev = self.get_pointer_device()
            result = ptr_dev.grab(
                self.window.get_window(),
                Gdk.GrabOwnership.NONE,
                True,
                Gdk.EventMask.ALL_EVENTS_MASK,
                None,
                0
            )
            logger.debug("Focus in event, grabbing pointer: %s", result)
        self.is_focused = True

    def on_input_changed(self, entry):
        """
        Triggered by user input
        """
        Search.get_instance().on_query_change(self._get_user_query())

    # pylint: disable=inconsistent-return-statements
    def on_input_key_press_event(self, widget, event):
        keyval = event.get_keyval()
        keyname = Gdk.keyval_name(keyval[1])
        alt = event.state & Gdk.ModifierType.MOD1_MASK
        ctrl = event.state & Gdk.ModifierType.CONTROL_MASK
        Search.get_instance().on_key_press_event(widget, event, self._get_user_query())

        if keyname == 'Escape':
            self.hide()

        elif ctrl and keyname == 'comma':
            self.activate_preferences()

        if self.results_nav:
            if keyname in ('Up', 'ISO_Left_Tab') or (ctrl and keyname == 'p'):
                self.results_nav.go_up()
                return True
            if keyname in ('Down', 'Tab') or (ctrl and keyname == 'n'):
                self.results_nav.go_down()
                return True
            if alt and keyname in ('Return', 'KP_Enter'):
                self.enter_result_item(alt=True)
            elif keyname in ('Return', 'KP_Enter'):
                self.enter_result_item()
            elif alt and keyname.isdigit() and 0 < int(keyname) < 10:
                # on Alt+<num>
                try:
                    self.enter_result_item(int(keyname) - 1)
                except IndexError:
                    # selected non-existing result item
                    pass
            elif alt and len(keyname) == 1 and 97 <= ord(keyname) <= 122:
                # on Alt+<char>
                try:
                    self.enter_result_item(ord(keyname) - 97 + 9)
                except IndexError:
                    # selected non-existing result item
                    pass

    ######################################
    # Helpers
    ######################################

    def get_input(self):
        return self.input

    def fix_window_width(self):
        """
        Add 2px to the window width if GTK+ >= 3.20
        Because of the bug in <3.20 that doesn't add css borders to the width
        """
        if gtk_version_is_gte(3, 20, 0):
            width, height = self.get_size_request()
            self.set_size_request(width + 2, height)

    def init_theme(self):
        load_available_themes()
        theme = Theme.get_current()
        theme.clear_cache()

        # removing window shadow solves issue with DEs without a compositor (#230)
        if get_options().no_window_shadow:
            self.window_body.get_style_context().add_class('no-window-shadow')

        self._render_prefs_icon()
        self.init_styles(theme.compile_css())

    def activate_preferences(self, page='preferences'):
        """
        From the PyGTK Reference manual
        Say for example the preferences dialog is currently open,
        and the user chooses Preferences from the menu a second time;
        use the present() method to move the already-open dialog
        where the user can see it.
        """
        self.hide()

        if self.preferences_dialog is not None:
            logger.debug('show existing preferences_dialog')
            self.preferences_dialog.present(page=page)
        elif self.PreferencesDialog is not None:
            logger.debug('create new preferences_dialog')
            self.preferences_dialog = self.PreferencesDialog()  # pylint: disable=E1102
            self.preferences_dialog.connect('destroy', self.on_preferences_dialog_destroyed)
            self.preferences_dialog.show(page=page)
        # destroy command moved into dialog to allow for a help button

    def position_window(self):
        window_width = self.get_size()[0]
        screen = get_current_screen_geometry()

        if self.settings.get_property('render-on-screen') == "default-monitor":
            screen = get_primary_screen_geometry()

        # The topmost pixel of the window should be at 1/5 of the current screen's height
        # Window should be positioned in the center horizontally
        # Also, add offset x and y, in order to move window to the current screen
        self.move(screen['width'] / 2 - window_width / 2 + screen['x'],
                  screen['height'] / 5 + screen['y'])

    def show_window(self):
        # works only when the following methods are called in that exact order
        self.window.set_sensitive(True)
        self.window.present()
        self.position_window()
        if not is_wayland_compatibility_on():
            self.present_with_time(Keybinder.get_current_event_time())

        if not self.input.get_text():
            # make sure frequent apps are shown if necessary
            self.show_results([])
        elif self.settings.get_property('clear-previous-query'):
            self.input.set_text('')
        else:
            self.input.grab_focus()

    def toggle_window(self, key=None):
        if self.is_visible():
            self.hide()
        else:
            self.show_window()

    def bind_show_app_hotkey(self, accel_name):
        if is_wayland_compatibility_on():
            return

        if self._current_accel_name == accel_name:
            return

        if self._current_accel_name:
            Keybinder.unbind(self._current_accel_name)
            self._current_accel_name = None

        logger.info("Trying to bind app hotkey: %s", accel_name)
        Keybinder.bind(accel_name, self.toggle_window)
        self._current_accel_name = accel_name
        self.notify_hotkey_change(accel_name)

    def notify_hotkey_change(self, accel_name):
        (key, mode) = Gtk.accelerator_parse(accel_name)
        display_name = Gtk.accelerator_get_label(key, mode)
        app_cache_db = AppCacheDb.get_instance()
        if not app_cache_db.find('startup_hotkey_notification'):
            app_cache_db.put('startup_hotkey_notification', True)
            app_cache_db.commit()
            show_notification("Ulauncher", "Hotkey is set to %s" % display_name)

    def _get_user_query(self):
        return Query(self.input.get_text())

    def select_result_item(self, index, onHover=False):
        if time.time() - self._results_render_time > 0.1:
            # Work around issue #23 -- don't automatically select item if cursor is hovering over it upon render
            self.results_nav.select(index)

    def enter_result_item(self, index=None, alt=False):
        if not self.results_nav.enter(self._get_user_query(), index, alt=alt):
            # hide the window if it has to be closed on enter
            self.hide_and_clear_input()

    def hide(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """Override the hide method to ensure the pointer grab is released."""
        if self.settings.get_property('grab-mouse-pointer'):
            self.get_pointer_device().ungrab(0)
        super().hide(*args, **kwargs)

    def get_pointer_device(self):
        return (self
                .window
                .get_window()
                .get_display()
                .get_device_manager()
                .get_client_pointer())

    def hide_and_clear_input(self):
        self.input.set_text('')
        self.hide()

    def show_results(self, result_items):
        """
        :param list result_items: list of ResultItem instances
        """
        self.results_nav = None
        self.result_box.foreach(lambda w: w.destroy())

        show_recent_apps = self.settings.get_property('show-recent-apps')
        recent_apps_number = 3 if show_recent_apps else 0
        try:
            recent_apps_number = int(str(show_recent_apps))
        except ValueError:
            pass
        if not result_items and not self.input.get_text() and recent_apps_number > 0:
            result_items = AppStatDb.get_instance().get_most_frequent(recent_apps_number)

        results = self.create_item_widgets(result_items, self._get_user_query())

        if results:
            self._results_render_time = time.time()
            for item in results:
                self.result_box.add(item)
            self.results_nav = ItemNavigation(self.result_box.get_children())
            self.results_nav.select_default(self._get_user_query())

            self.result_box.show_all()
            self.result_box.set_margin_bottom(10)
            self.result_box.set_margin_top(3)
            self.apply_css(self.result_box)
        else:
            self.result_box.set_margin_bottom(0)
            self.result_box.set_margin_top(0)
        logger.debug('render %s results', len(results))

    def _render_prefs_icon(self):
        scale_factor = get_monitor_scale_factor()
        prefs_pixbuf = load_image(get_data_file('media', 'gear.svg'), 16 * scale_factor)
        surface = Gdk.cairo_surface_create_from_pixbuf(prefs_pixbuf, scale_factor, self.get_window())
        prefs_image = Gtk.Image.new_from_surface(surface)
        self.prefs_btn.set_image(prefs_image)

    @staticmethod
    def create_item_widgets(items, query):
        results = []
        for index, result_item in enumerate(items):
            glade_filename = get_data_file('ui', '%s.ui' % result_item.UI_FILE)
            if not os.path.exists(glade_filename):
                glade_filename = None

            builder = Gtk.Builder()
            builder.set_translation_domain('ulauncher')
            builder.add_from_file(glade_filename)

            item_frame = builder.get_object('item-frame')
            item_frame.initialize(builder, result_item, index, query)

            results.append(item_frame)

        return results



import gi
# pylint: disable=wrong-import-position
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
from gi.repository import Gtk, Gdk

from ulauncher.api.shared.action.BaseAction import BaseAction


class CopyToClipboardAction(BaseAction):
    """
    Copy text to the clipboard

    :param str text: text to copy
    """

    def __init__(self, text):
        self.text = text

    def keep_app_open(self):
        return False

    def run(self):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(self.text, -1)
        clipboard.store()



import logging
from typing import Any
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
# pylint: disable=wrong-import-position
from gi.repository import Gtk, Gdk

from ulauncher.utils.Theme import Theme
from ulauncher.utils.display import get_monitor_scale_factor
from ulauncher.search.Query import Query

logger = logging.getLogger(__name__)


# type: ignore
class ResultItemWidget(Gtk.EventBox):
    __gtype_name__ = "ResultItemWidget"

    shortcut = ''  # type: str
    index = 0  # type: int
    builder = None  # type: Any
    name = ''  # type: str
    query = Query('')  # type: Query
    item_object = None  # type: Any
    item_box = None  # type: Any

    def initialize(self, builder: Any, item_object: Any, index: int, query: Query) -> None:
        self.builder = builder
        item_frame = self.builder.get_object('item-frame')
        item_frame.connect("button-release-event", self.on_click)
        item_frame.connect("enter_notify_event", self.on_mouse_hover)

        self.item_box = builder.get_object('item-box')
        self.item_object = item_object
        self.query = query
        self.set_index(index)

        self.set_icon(item_object.get_icon())
        self.set_description(item_object.get_description(query))
        self.set_name_highlighted()

    def set_index(self, index):
        """
        Set index for the item and assign shortcut
        """
        self.index = index
        # Alt+1..9, then Alt+a..z
        index_text = index + 1 if index < 9 else chr(97 + index - 9)
        self.shortcut = 'Alt+%s' % index_text
        self.set_shortcut(self.shortcut)

    def select(self):
        self.set_name_highlighted(True)
        self.item_box.get_style_context().add_class('selected')

    def deselect(self):
        self.set_name_highlighted(False)
        self.item_box.get_style_context().remove_class('selected')

    def set_icon(self, icon):
        """
        :param PixBuf icon:
        """
        if not icon:
            return

        iconWgt = self.builder.get_object('item-icon')
        scale_factor = get_monitor_scale_factor()

        if scale_factor == 1:
            iconWgt.set_from_pixbuf(icon)
            return

        try:
            surface = Gdk.cairo_surface_create_from_pixbuf(icon, scale_factor, self.get_window())
            iconWgt.set_from_surface(surface)
        except AttributeError:  # Fallback for GTK+ older than 3.10
            iconWgt.set_from_pixbuf(icon)

    def set_name_highlighted(self, is_selected: bool = False) -> None:
        colors = Theme.get_current().get('matched_text_hl_colors')
        color = colors['when_selected'] if is_selected else colors['when_not_selected']
        self.set_name(self.item_object.get_name_highlighted(self.query, color) or self.item_object.get_name())

    # pylint: disable=arguments-differ
    def set_name(self, name: str) -> None:
        item = self.builder.get_object('item-name')
        if '<span' in name:  # dealing with markup
            item.set_markup(name)
        else:
            item.set_text(name)
        self.name = name

    def get_name(self):
        return self.name

    # pylint: disable=unused-argument
    def on_click(self, widget, event=None):
        self.get_toplevel().select_result_item(self.index)
        alt_enter = bool(event and event.button != 1)
        self.get_toplevel().enter_result_item(alt=alt_enter)

    def on_mouse_hover(self, widget, event):
        self.get_toplevel().select_result_item(self.index, onHover=True)

    def set_description(self, description):
        description_obj = self.builder.get_object('item-descr')
        if not description_obj:
            return

        if description:
            description_obj.set_text(description)
        else:
            description_obj.destroy()  # remove description label
            self.builder.get_object('item-name').set_margin_top(8)  # shift name label down to the center

    def set_shortcut(self, text):
        self.builder.get_object('item-shortcut').set_text(text)

    def on_enter(self, query):
        return self.item_object.on_enter(query)

    def on_alt_enter(self, query):
        return self.item_object.on_alt_enter(query)

    def get_keyword(self):
        return self.item_object.get_keyword()

    def selected_by_default(self, query):
        return self.item_object.selected_by_default(query)



"""
Module for detecting different states with respect to Wayland.

These functions are in a standalone module to minimize the dependencies so that
ulauncher.utils.xinit can safely use it.
"""
import os


def is_wayland():
    return os.environ.get('XDG_SESSION_TYPE', '').lower() == 'wayland'


def is_wayland_compatibility_on():
    """
    In this mode user won't be able to set app hotkey via preferences
    Set hotkey in OS Settings > Devices > Keyboard > Add Hotkey > Command: ulauncher-toggle
    GDK_BACKEND is typically unset in Wayland sessions to allow GTK apps to self-select
    """
    return is_wayland() and (gdk_backend() == '' or gdk_backend().lower().startswith('wayland'))


def gdk_backend():
    return os.environ.get('GDK_BACKEND', '')



import re

_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')


def split_camel_case(text: str, sep: str = '_') -> str:
    s1 = _first_cap_re.sub(r'\1%s\2' % sep, text)
    return _all_cap_re.sub(r'\1%s\2' % sep, s1).lower()



"""
Simple module to call XInitThreads() to properly support multi-threaded X11 applications.
"""
from ulauncher.utils.wayland import is_wayland_compatibility_on


if not is_wayland_compatibility_on():
    from ctypes import cdll

    # Using libX11.so.6 may seem a bit hard-coded, but a quick search on the Internet indicates
    # discussion of this ABI version back to before 2009, 12 years prior to when this code was added.
    # Also, with most of the development focus on Wayland, the chance of a ABI version bump in X11 is
    # exceptionally small. Therefore, no additional fancy discovery methods are necessary.
    x11 = cdll.LoadLibrary("libX11.so.6")

    # In order to use any X11 calls from multiple threads, the X11 library must be initialized to use
    # mutex protection. XInitThreads() sets this up and must be called before any other X11
    # initialization is performed. Since this always needs to be called before other imports anyway,
    # this module does the init at module import time.
    x11.XInitThreads()



from threading import Timer


def debounce(wait):
    """ Decorator that will postpone a functions
        execution until after wait seconds
        have elapsed since the last time it was invoked. """
    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)

            try:
                debounced.t.cancel()
            except AttributeError:
                pass

            debounced.t = Timer(wait, call_it)
            debounced.t.start()

        return debounced

    return decorator



import logging
from functools import partial
from ulauncher.api.server.port_finder import find_unused_port
from ulauncher.api.server.ExtensionController import ExtensionController
from ulauncher.utils.SimpleWebSocketServer import SimpleWebSocketServer
from ulauncher.utils.decorator.run_async import run_async
from ulauncher.utils.decorator.singleton import singleton

logger = logging.getLogger(__name__)


class ExtensionServer:

    @classmethod
    @singleton
    def get_instance(cls):
        return cls()

    def __init__(self):
        self.hostname = '127.0.0.1'
        self.port = None
        self.ws_server = None
        self.controllers = {}

    def generate_ws_url(self, extension_id):
        """
        Returns WebSocket URL for given `extension_id`

        :rtype: str
        """
        if not self.is_running():
            raise ServerIsNotRunningError()

        return 'ws://%s:%s/%s' % (self.hostname, self.port, extension_id)

    def start(self):
        """
        Starts WS server
        """
        if self.ws_server:
            raise ServerIsRunningError()

        self._start_thread()

    @run_async(daemon=True)
    def _start_thread(self):
        self.port = self.port or find_unused_port(5054)
        logger.info('Starting WS server on port %s', self.port)
        self.ws_server = SimpleWebSocketServer(self.hostname,
                                               self.port,
                                               partial(ExtensionController, self.controllers))
        self.ws_server.serveforever()
        self.ws_server = None
        logger.warning('WS server exited')

    def stop(self):
        """
        Stops WS server
        """
        if not self.is_running():
            raise ServerIsNotRunningError()

        self.ws_server.close()

    def is_running(self):
        """
        :rtype: bool
        """
        return bool(self.ws_server)

    def get_controller(self, extension_id):
        """
        :param str extension_id:
        :rtype: ~ulauncher.api.server.ExtensionController.ExtensionController
        """
        return self.controllers.get(extension_id)

    def get_controllers(self):
        """
        :rtype: list of  :class:`~ulauncher.api.server.ExtensionController.ExtensionController`
        """
        return self.controllers.values()

    def get_controller_by_keyword(self, keyword):
        """
        :param str keyword:
        :rtype: ~ulauncher.api.server.ExtensionController.ExtensionController
        """
        for _, ctl in self.controllers.items():
            if keyword in ctl.preferences.get_active_keywords():
                return ctl

        return None


class ServerIsRunningError(RuntimeError):
    pass


class ServerIsNotRunningError(RuntimeError):
    pass


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    server = ExtensionServer.get_instance()
    server.start()



from ulauncher.api.shared.action.CopyToClipboardAction import CopyToClipboardAction
from ulauncher.api.shared.action.DoNothingAction import DoNothingAction
from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.config import get_data_file
from ulauncher.utils.image_loader import load_image


class CalcResultItem(ResultItem):

    # pylint: disable=super-init-not-called
    def __init__(self, result: str = None, error: str = 'Unknown error'):
        self.result = result
        self.error = error

    def get_name(self) -> str:
        return str(self.result) if self.result is not None else 'Error!'

    # pylint: disable=super-init-not-called, arguments-differ
    def get_name_highlighted(self, *args) -> None:
        pass

    def get_description(self, query) -> str:
        return 'Enter to copy to the clipboard' if self.result else self.error

    def get_icon(self):
        return load_image(get_data_file('media/calculator-icon.png'), self.get_icon_size())

    def on_enter(self, query):
        if self.result is not None:
            return CopyToClipboardAction(str(self.result))

        return DoNothingAction()



from typing import Dict, Callable, Any

CachedInstances = Dict[Callable, Any]
objects = {}  # type: CachedInstances


def singleton(fn):
    """
    Decorator function.
    Call to a decorated function always returns the same instance

    Note: it doesn't take into account args and kwargs when looks up a saved instance
    Call a decorated function with `spawn=True` in order to get a new instance
    """
    def wrapper(*args, **kwargs):
        if not kwargs.get('spawn') and objects.get(fn):
            return objects[fn]

        instance = fn(*args, **kwargs)
        objects[fn] = instance
        return instance

    return wrapper



from ulauncher.api.shared.action.CopyToClipboardAction import CopyToClipboardAction
from ulauncher.api.shared.item.SmallResultItem import SmallResultItem
from ulauncher.utils.image_loader import get_themed_icon_by_name


class CopyPathToClipboardItem(SmallResultItem):
    """
    :param ~ulauncher.utils.Path.Path path:
    """

    # pylint: disable=super-init-not-called
    def __init__(self, path):
        self.path = path

    def get_name(self):
        return 'Copy Path to Clipboard'

    # pylint: disable=super-init-not-called, arguments-differ
    def get_name_highlighted(self, *args):
        pass

    def get_icon(self):
        return get_themed_icon_by_name('edit-copy', self.get_icon_size())

    def on_enter(self, query):
        return CopyToClipboardAction(self.path.get_abs_path())



import os
import json
import gi
gi.require_version('GObject', '2.0')
# pylint: disable=wrong-import-position
from gi.repository import GObject

from ulauncher.utils.decorator.singleton import singleton
from ulauncher.config import SETTINGS_FILE_PATH

DEFAULT_BLACKLISTED_DIRS = [
    '/usr/share/locale',
    '/usr/share/app-install',
    '/usr/share/kservices5',
    '/usr/share/fk5',
    '/usr/share/kservicetypes5',
    '/usr/share/applications/screensavers',
    '/usr/share/kde4',
    '/usr/share/mimelnk'
]

GPROPERTIES = {
    "hotkey-show-app": (str,  # type
                        "Hotkey: Show ulauncher window",  # nick name
                        None,  # description
                        "<Primary>space",  # default
                        GObject.ParamFlags.READWRITE),  # flags
    "show-indicator-icon": (bool,
                            "Show app indicator",
                            None,
                            True,
                            GObject.ParamFlags.READWRITE),
    "show-recent-apps": (str,
                         "Number of recent apps",
                         None,
                         "0",
                         GObject.ParamFlags.READWRITE),
    "clear-previous-query": (bool,
                             "Clear query when app looses focus",
                             None,
                             True,
                             GObject.ParamFlags.READWRITE),
    "theme-name": (str,
                   "Current theme",
                   None,
                   "light",
                   GObject.ParamFlags.READWRITE),
    "blacklisted-desktop-dirs": (str,
                                 "Blacklisted desktop dirs",
                                 None,
                                 ':'.join(DEFAULT_BLACKLISTED_DIRS),
                                 GObject.ParamFlags.READWRITE),
    "render-on-screen": (str,
                         "Monitor to render on",
                         None,
                         "mouse-pointer-monitor",
                         GObject.ParamFlags.READWRITE),
    "terminal-command": (str,
                         "Terminal command",
                         None,
                         "",
                         GObject.ParamFlags.READWRITE),
    "grab-mouse-pointer": (bool,
                           "Grab mouse while open (prevents losing focus)",
                           None,
                           False,
                           GObject.ParamFlags.READWRITE),
    "disable-desktop-filters": (bool,  # type
                                "Display all apps in environment despite OnlyShowIn/NotShowIn",  # nick name
                                None,  # description
                                False,  # default
                                GObject.ParamFlags.READWRITE),  # flags
}


class Settings(GObject.GObject):
    """
    Get/Set properties using :code:`settings.get_property('prop-name')` or
    :code:`settings.set_property('prop-name', 'new value')`

    Subscribe to property changes:

    >>> def on_notify(settings, prop):
    ...     print(prop.name)
    >>>
    >>> settings.connect("notify::hotkey-show-app", on_notify)
    """

    # __gproperties__ is used only to register properties
    # More info about __gproperties__
    # https://www.pygtk.org/articles/subclassing-gobject/sub-classing-gobject-in-python.htm#d0e127
    # https://python-gtk-3-tutorial.readthedocs.org/en/latest/objects.html
    __gproperties__ = GPROPERTIES

    _filename = None
    _properties = None

    @classmethod
    @singleton
    def get_instance(cls):
        settings = cls()
        settings.load_from_file(SETTINGS_FILE_PATH)
        return settings

    def __init__(self):
        GObject.GObject.__init__(self)
        self._filename = None
        self._properties = dict((name, opts[-2]) for name, opts in GPROPERTIES.items())

    def load_from_file(self, filename):
        self._filename = filename
        if os.path.exists(filename):
            if not os.path.isfile(filename):
                raise IOError("%s exists and is not a file" % filename)

            with open(filename, 'r') as f:
                # Convert underscore to dash, in case user migrated to v6, saved settings and reverted
                self._properties = {key.replace("_", "-"): val for key, val in json.load(f).items()}
        else:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            self.save_to_file()

    def save_to_file(self):
        if not self._filename:
            return

        with open(self._filename, 'w') as f:
            json.dump(self._properties, f, indent=4, sort_keys=True)

    def do_get_property(self, prop):
        try:
            return self._properties[prop.name]
        except KeyError:
            # return default
            return GPROPERTIES[prop.name][3]

    def do_set_property(self, prop, value):
        self._properties[prop.name] = value



from functools import wraps
import logging

import gi
gi.require_version('Gtk', '3.0')

# AppIndicator support is optional. It'll work if you install
# gir1.2-ayatanaappindicator3-0.1 or an equivalent package for your distro
# pylint: disable=wrong-import-position
appIndicatorSupported = False
try:
    gi.require_version('AppIndicator3', '0.1')
    appIndicatorSupported = True
except ValueError:
    pass

try:
    gi.require_version('AyatanaAppIndicator3', '0.1')
    appIndicatorSupported = True
except ValueError:
    pass

try:
    from gi.repository import AppIndicator3
except ImportError:
    pass

try:
    from gi.repository import AyatanaAppIndicator3 as AppIndicator3  # type: ignore # noqa: F811
except ImportError:
    pass

from gi.repository import Gtk
from ulauncher.utils.decorator.singleton import singleton

logger = logging.getLogger(__name__)


def onlyIfAppindicatorIsSupported(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        if appIndicatorSupported:
            return func(*args, **kwargs)
        return None
    return wrapped


class AppIndicator:

    @classmethod
    @singleton
    def get_instance(cls):
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow
        window = UlauncherWindow.get_instance()
        indicator = cls("ulauncher")
        indicator.set_icon('ulauncher-indicator')
        indicator.add_menu_item(window.on_mnu_preferences_activate, "Preferences")
        indicator.add_menu_item(window.on_mnu_about_activate, "About")
        indicator.add_separator()
        indicator.add_menu_item(Gtk.main_quit, "Exit")
        return indicator

    def __init__(self, iconname):
        if appIndicatorSupported:
            self.__menu = Gtk.Menu()
            self.__indicator = AppIndicator3.Indicator.new(
                iconname,
                "",
                AppIndicator3.IndicatorCategory.APPLICATION_STATUS
            )
            self.__indicator.set_menu(self.__menu)

    @onlyIfAppindicatorIsSupported
    def set_icon(self, path):
        self.__indicator.set_icon(path)

    @onlyIfAppindicatorIsSupported
    def switch(self, on=False):
        if on:
            self.show()
        else:
            self.hide()

    @onlyIfAppindicatorIsSupported
    def add_menu_item(self, command, title):
        menu_item = Gtk.MenuItem()
        menu_item.set_label(title)
        menu_item.connect("activate", command)
        self.__menu.append(menu_item)

    @onlyIfAppindicatorIsSupported
    def add_separator(self):
        menu_item = Gtk.SeparatorMenuItem()
        self.__menu.append(menu_item)

    @onlyIfAppindicatorIsSupported
    def show(self):
        self.__indicator.set_status(AppIndicator3.IndicatorStatus.ACTIVE)
        self.__menu.show_all()

    @onlyIfAppindicatorIsSupported
    def hide(self):
        self.__indicator.set_status(AppIndicator3.IndicatorStatus.PASSIVE)

    @onlyIfAppindicatorIsSupported
    def right_click_event_statusicon(self, icon, button, time):
        self._get_tray_menu()

        def pos(menu, aicon):
            return Gtk.StatusIcon.position_menu(menu, aicon)

        self.__menu.popup(None, None, pos, icon, button, time)

    def _get_tray_menu(self):
        return self.__menu



import logging
from functools import partial
from threading import Timer
import gi
gi.require_version('GLib', '2.0')
# pylint: disable=wrong-import-position
from gi.repository import GLib

from ulauncher.api.shared.action.DoNothingAction import DoNothingAction
from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.utils.decorator.singleton import singleton

logger = logging.getLogger('DeferredResultRenderer')


class DeferredResultRenderer:
    """
    Handles asynchronous render for extensions
    """

    LOADING_DELAY = 0.3  # delay in sec before Loading... is rendered

    @classmethod
    @singleton
    def get_instance(cls) -> 'DeferredResultRenderer':
        """
        Returns singleton instance
        """
        return cls()

    def __init__(self):
        self.loading = None
        self.active_event = None
        self.active_controller = None

    def get_active_controller(self):
        return self.active_controller

    def handle_event(self, event, controller):
        """
        Schedules "Loading..." message

        :rtype: :class:`~ulauncher.api.shared.action.DoNothingAction.DoNothingAction`
        """
        self._cancel_loading()
        self.loading = Timer(self.LOADING_DELAY,
                             partial(self._render_loading,
                                     controller.get_manifest().load_icon(ResultItem.get_icon_size())))
        self.loading.start()
        self.active_event = event
        self.active_controller = controller

        return DoNothingAction()

    def handle_response(self, response, controller):
        """
        Calls :func:`response.action.run`
        """
        if self.active_controller != controller or self.active_event != response.event:
            return

        self._cancel_loading()
        response.action.run()
        if not response.action.keep_app_open():
            self._hide_window()

    def on_query_change(self):
        """
        Cancel "Loading...", reset active_event and active_controller
        """
        self._cancel_loading()
        self.active_event = None
        self.active_controller = None

    def _cancel_loading(self):
        if self.loading:
            self.loading.cancel()
            self.loading = None

    def _render_loading(self, icon):
        loading_item = ResultItem(name='Loading...',
                                  icon=icon,
                                  highlightable=False)
        RenderResultListAction([loading_item]).run()

    def _hide_window(self):
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow

        window = UlauncherWindow.get_instance()
        if window.is_visible():
            # update UI in the main thread to avoid race conditions
            GLib.idle_add(window.hide_and_clear_input)



import socket


PORT_RANGE = list(range(5001, 5100))


def find_unused_port(default=5001):
    """
    Returns first free port in range :data:`PORT_RANGE`

    :param int default:
    :rtype: int
    """
    to_select = [default] + PORT_RANGE
    for port in to_select:
        if not is_port_in_use(port):
            return port

    raise Exception("Could not find an unused port")


def is_port_in_use(port, host='127.0.0.1'):
    """
    :rtype bool:
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((host, port))
    return result == 0



import os
import sys
import pickle
import logging
import traceback
from threading import Timer

import websocket
from ulauncher.api.shared.event import SystemExitEvent

logger = logging.getLogger(__name__)


class Client:
    """
    Instantiated in extension code and manages data transfer from/to Ulauncher app

    :param ~ulauncher.api.client.Extension extension:
    :param str ws_api_url: uses env. var `ULAUNCHER_WS_API` by default
    """

    def __init__(self, extension, ws_api_url=None):
        self.ws_api_url = ws_api_url or os.environ.get('ULAUNCHER_WS_API')
        if not self.ws_api_url:
            raise Exception('ULAUNCHER_WS_API was not specified')

        self.extension = extension
        self.ws = None

    def connect(self):
        """
        Connects to WS server and blocks thread
        """
        websocket.enableTrace(False)
        # pylint: disable=unnecessary-lambda
        self.ws = websocket.WebSocketApp(self.ws_api_url,
                                         on_message=lambda ws, msg: self.on_message(ws, msg),
                                         on_error=lambda ws, error: self.on_error(ws, error),
                                         on_open=lambda ws: self.on_open(ws),
                                         on_close=lambda ws: self.on_close(ws))
        self.ws.run_forever()

    # pylint: disable=unused-argument
    def on_message(self, ws, message):
        """
        Parses message from Ulauncher and triggers extension event

        :param websocket.WebSocketApp ws:
        :param str message:
        """
        event = pickle.loads(message)
        logger.debug('Incoming event %s', type(event).__name__)
        try:
            self.extension.trigger_event(event)
        # pylint: disable=broad-except
        except Exception:
            traceback.print_exc(file=sys.stderr)

    def on_error(self, ws, error):
        logger.error('WS Client error %s', error)

    def on_close(self, ws):
        """
        Terminates extension process on WS disconnect.

        Triggers :class:`~ulauncher.api.shared.event.SystemExitEvent` for graceful shutdown

        :param websocket.WebSocketApp ws:
        """
        logger.warning("Connection closed. Exiting")
        self.extension.trigger_event(SystemExitEvent())
        # extension has 0.5 sec to save it's state, after that it will be terminated
        Timer(0.5, os._exit, args=[0]).start()

    def on_open(self, ws):
        self.ws = ws

    def send(self, response):
        """
        Sends response to Ulauncher

        :param ~ulauncher.api.shared.Response.Response response:
        """
        logger.debug('Send message')
        self.ws.send(pickle.dumps(response))



import os
import logging
from functools import lru_cache

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Gio', '2.0')
gi.require_version('GdkPixbuf', '2.0')

# pylint: disable=wrong-import-position
from gi.repository import Gtk, Gio, GLib, GdkPixbuf
from ulauncher.config import get_data_file


# type: ignore
icon_theme = Gtk.IconTheme.get_default()
logger = logging.getLogger(__name__)


@lru_cache(maxsize=50)
def load_image(path, size):
    """
    :param str path:
    :param int size:

    :rtype: :class:`GtkPixbuf` or :code:`None`
    :returns: None if :func:`new_from_file_at_size` raises error
    """
    path = os.path.expanduser(path)
    # pylint: disable=broad-except
    try:
        return GdkPixbuf.Pixbuf.new_from_file_at_size(path, size, size)
    except Exception as e:
        logger.warning('Could not load image %s. E: %s', path, e)


def get_app_icon_pixbuf(icon, icon_size, icon_name):
    """
    :param Gio.Icon icon:
    :param int icon_size:
    :param str icon_name:
    :rtype: :class:`GtkPixbuf`
    """
    pixbuf_icon = None
    # pylint: disable=broad-except

    if isinstance(icon, Gio.ThemedIcon):
        try:
            icon_name = icon.get_names()[0]
            if not icon_name:
                return None
            pixbuf_icon = get_themed_icon_by_name(icon_name, icon_size)
        except Exception as e:
            logger.info('Could not load icon for %s. E: %s', icon_name, e)

    elif isinstance(icon, Gio.FileIcon):
        pixbuf_icon = load_image(icon.get_file().get_path(), icon_size)

    elif isinstance(icon, str):
        pixbuf_icon = load_image(icon, icon_size)

    if not pixbuf_icon:
        pixbuf_icon = load_image(get_data_file('media', 'executable-icon.png'), icon_size)

    return pixbuf_icon


def check_icon_name(icon_name):
    if icon_theme.has_icon(icon_name):
        return icon_name
    for icon in icon_theme.list_icons():
        if icon.startswith(icon_name):
            logger.warning('Icon theme is broken or incomplete. Substituting "%s" for "%s"', icon_name, icon)
            return icon
    if icon_name != "application-x-executable":
        logger.warning('Icon "%s" completely missing from icon theme. Consider switching icon theme.', icon_name)
        logger.info('Falling back on default icon "application-x-executable"')
        return check_icon_name("application-x-executable")


def get_themed_icon_by_name(icon_name, icon_size):
    _icon_name = check_icon_name(icon_name)
    if _icon_name:
        return icon_theme.load_icon(_icon_name, icon_size, Gtk.IconLookupFlags.FORCE_SIZE)
    return load_image(get_data_file('media', 'unknown-file-icon.png'), icon_size)


ULAUNCHER_FILE_ICON_DB = ['3g2', '3gp', 'ai', 'air', 'asf', 'avi', 'bib', 'cls', 'csv', 'deb', 'djvu', 'dmg', 'doc',
                          'docx', 'dwf', 'dwg', 'eps', 'epub', 'exe', 'f77', 'f90', 'f', 'flac', 'flv', 'gif', 'gz',
                          'ico', 'indd', 'iso', 'jpeg', 'jpg', 'log', 'm4a', 'm4v', 'midi', 'mkv', 'mov', 'mp3', 'mp4',
                          'mpeg', 'mpg', 'msi', 'odp', 'ods', 'odt', 'oga', 'ogg', 'ogv', 'pdf', 'png', 'pps', 'ppsx',
                          'ppt', 'pptx', 'psd', 'pub', 'py', 'qt', 'ra', 'ram', 'rar', 'rm', 'rpm', 'rtf', 'rv', 'skp',
                          'spx', 'sql', 'sty', 'tar', 'tex', 'tgz', 'tiff', 'ttf', 'txt', 'vob', 'wav', 'wmv', 'xls',
                          'xlsx', 'xml', 'xpi', 'zip']

SPECIAL_DIRS = {
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOWNLOAD): 'folder-download',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOCUMENTS): 'folder-documents',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_MUSIC): 'folder-music',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PICTURES): 'folder-pictures',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PUBLIC_SHARE): 'folder-publicshare',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_TEMPLATES): 'folder-templates',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_VIDEOS): 'folder-videos',
    GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP): 'user-desktop',
    os.path.expanduser('~'): 'folder-home'
}

FREEDESKTOP_STANDARD = {
    'html': 'text-html'
}


def get_file_icon(path, icon_size):
    """
    :param ~ulauncher.utils.Path.Path path:
    :param int icon_size:
    """
    # pylint: disable=broad-except
    try:
        if path.is_dir():
            special_dir = SPECIAL_DIRS.get(str(path))
            if special_dir:
                return get_themed_icon_by_name(special_dir, icon_size)
            return get_themed_icon_by_name('folder', icon_size)

        ext = path.get_ext()
        if ext in ULAUNCHER_FILE_ICON_DB:
            return load_image(get_data_file('media', 'fileicons', '%s.png' % ext), icon_size)

        freedesktop = FREEDESKTOP_STANDARD.get(ext)
        if freedesktop:
            return get_themed_icon_by_name(freedesktop, icon_size)

        if path.is_exe():
            return load_image(get_data_file('media', 'executable-icon.png'), icon_size)
    except Exception as e:
        logger.warning('Icon not found %s. %s: %s', path, type(e).__name__, e)

    return load_image(get_data_file('media', 'unknown-file-icon.png'), icon_size)



import logging
import os
import re
import shlex
from shutil import which
from pathlib import Path
import gi

gi.require_version("GLib", "2.0")
# pylint: disable=wrong-import-position
from gi.repository import GLib

from ulauncher.utils.desktop.reader import read_desktop_file
from ulauncher.utils.Settings import Settings
from ulauncher.api.shared.action.BaseAction import BaseAction

logger = logging.getLogger(__name__)
settings = Settings.get_instance()
has_systemd = which("systemctl") and which("systemd-run")
runs_in_systemd = has_systemd and os.system('systemctl --user is-active --quiet ulauncher') == 0


class LaunchAppAction(BaseAction):
    """
    Launches app by given `.desktop` file path

    :param str filename: path to .desktop file
    """

    def __init__(self, filename):
        self.filename = filename

    def keep_app_open(self):
        return False

    def run(self):
        app = read_desktop_file(self.filename)
        app_id = Path(app.get_id()).stem
        app_exec = app.get_string('Exec')
        if app.get_boolean('DBusActivatable'):
            # https://wiki.gnome.org/HowDoI/DBusApplicationLaunching
            exec = ['gapplication', 'launch', app_id]
        elif app_exec:
            # strip field codes %f, %F, %u, %U, etc
            stripped_app_exec = re.sub(r'\%[uUfFdDnNickvm]', '', app_exec).rstrip()
            if app.get_boolean('Terminal'):
                terminal_exec = settings.get_property('terminal-command')
                if terminal_exec:
                    logger.info('Will run command in preferred terminal (%s)', terminal_exec)
                    exec = shlex.split(terminal_exec) + [stripped_app_exec]
                else:
                    exec = ['gtk-launch', app_id]
            else:
                exec = shlex.split(stripped_app_exec)

        if not exec:
            logger.error("No command to run %s", app_id)
        else:
            if runs_in_systemd and not app.get_boolean('X-Ulauncher-Inherit-Scope'):
                logger.warning("Will attempt to launch the app using systemd-run with --scope argument")
                logger.warning("This prevents the apps from terminating if Ulauncher crashes or is restarted.")
                logger.warning("On some systems with outdated systemd or incorrect permissions this doesn't work.")
                logger.warning("If this happens to you, don't run Ulauncher from systemd.")
                exec = ['systemd-run', '--user', '--scope'] + exec

            env = dict(os.environ.items())
            # Make sure GDK apps aren't forced to use x11 on wayland due to ulauncher's need to run
            # under X11 for proper centering.
            env.pop("GDK_BACKEND", None)

            try:
                logger.info('Run application %s (%s) Exec %s', app.get_name(), self.filename, exec)
                envp = ["{}={}".format(k, v) for k, v in env.items()]
                GLib.spawn_async(
                    argv=exec,
                    envp=envp,
                    flags=GLib.SpawnFlags.SEARCH_PATH_FROM_ENVP | GLib.SpawnFlags.SEARCH_PATH,
                    # setsid is really only needed if systemd-run is missing, but doesn't hurt to have.
                    child_setup=os.setsid
                )
            except Exception as e:
                logger.error('%s: %s', type(e).__name__, e)



import re

from ulauncher.api.shared.errors import UlauncherAPIError


class ProcessErrorExtractor:
    def __init__(self, error: str):
        """
        expecting an error like this:
            ModuleNotFoundError: No module named 'mymodule'
        """
        self.error = error

    def is_import_error(self) -> bool:
        return 'ModuleNotFoundError' in self.error

    def get_missing_package_name(self) -> str:
        """
        Returns a name of a module that extension failed to import
        """
        match = re.match(r"^.*'(\w+)'$", self.error)
        if not match:
            raise UlauncherAPIError('Could not extract errored module name from process output')

        return match.group(1)



from typing import Any, Dict

from ulauncher.utils.mypy_extensions import TypedDict
from ulauncher.utils.decorator.singleton import singleton
from ulauncher.utils.image_loader import get_app_icon_pixbuf
from ulauncher.search.apps.AppResultItem import AppResultItem


CachedItem = TypedDict('CachedItem', {
    'icon': Any,
    'name': str,
    'sizes': Dict[int, Any]
})


class AppIconCache:

    _icons = None  # type: Dict[str, CachedItem]

    @classmethod
    @singleton
    def get_instance(cls):
        return cls()

    def __init__(self):
        self._icons = {}

    def add_icon(self, desktop_file: str, icon: Any, name: str):
        """
        :param str desktop_file:
        :param Gio.Icon icon:
        :param str name:
        """
        self._icons[desktop_file] = {
            'icon': icon,
            'name': name,
            'sizes': {}
        }

    def get_pixbuf(self, desktop_file: str):
        """
        :param str desktop_file:
        :rtype: :class:`GtkPixbuf`
        """
        icon = self._icons.get(desktop_file)
        if not icon:
            return None

        size = AppResultItem.get_icon_size()
        try:
            return icon['sizes'][size]
        except KeyError:
            pass

        pixbuf = get_app_icon_pixbuf(icon['icon'], size, icon['name'])
        icon['sizes'][size] = pixbuf

        return pixbuf

    def remove_icon(self, desktop_file: str):
        if desktop_file in self._icons:
            del self._icons[desktop_file]



import html

from ulauncher.api.server.DeferredResultRenderer import DeferredResultRenderer
from ulauncher.api.server.ExtensionServer import ExtensionServer
from ulauncher.search.BaseSearchMode import BaseSearchMode
from ulauncher.api.server.ExtensionKeywordResultItem import ExtensionKeywordResultItem


class ExtensionSearchMode(BaseSearchMode):

    def __init__(self):
        self.extensionServer = ExtensionServer.get_instance()
        self.deferredResultRenderer = DeferredResultRenderer.get_instance()

    def is_enabled(self, query):
        """
        :param ~ulauncher.search.Query.Query query:
        :rtype: `True` if mode should be enabled for a query
        """
        return bool(self._get_controller_by_keyword(query.get_keyword())) and query.is_mode_active()

    def on_query_change(self, query):
        """
        Triggered when user changes a search query

        :param ~ulauncher.search.Query.Query query:
        """
        self.deferredResultRenderer.on_query_change()

    def handle_query(self, query):
        """
        :param ~ulauncher.search.Query.Query query:
        :rtype: :class:`~ulauncher.api.shared.action.BaseAction.BaseAction`
        """
        controller = self._get_controller_by_keyword(query.get_keyword())

        if not controller:
            raise Exception("Controller not found. (This line shouldn't be entered)")

        return controller.handle_query(query)

    def _get_controller_by_keyword(self, kw):
        """
        :param str kw: Keyword
        """
        return self.extensionServer.get_controller_by_keyword(kw)

    def get_searchable_items(self):
        """
        :rtype: list of :class:`~ulauncher.api.shared.item.ResultItem.ResultItem`
        """
        items = []
        for c in self.extensionServer.get_controllers():
            for pref in c.preferences.get_items(type='keyword'):
                if not pref['value']:
                    continue

                icon = c.manifest.load_icon(ExtensionKeywordResultItem.get_icon_size())
                items.append(ExtensionKeywordResultItem(name=html.escape(pref['name']),
                                                        description=html.escape(pref['description']),
                                                        keyword=pref['value'],
                                                        icon=icon))

        return items



import gi
gi.require_version('Gtk', '3.0')
# pylint: disable=wrong-import-position
from gi.repository import Gtk


class WindowHelper:

    css_provider = None

    def init_styles(self, path):
        if not self.css_provider:
            self.css_provider = Gtk.CssProvider()
        self.css_provider.load_from_path(path)
        self.apply_css(self)
        # pylint: disable=no-member
        visual = self.get_screen().get_rgba_visual()
        if visual:
            self.set_visual(visual)

    def apply_css(self, widget):
        Gtk.StyleContext.add_provider(widget.get_style_context(),
                                      self.css_provider,
                                      Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        if isinstance(widget, Gtk.Container):
            widget.forall(self.apply_css)



import logging
import os
import pickle  # Pickle is bad for you. This has been replaced with json in v6
from typing import Dict, TypeVar, Generic, Optional

Key = TypeVar('Key')
Value = TypeVar('Value')
Records = Dict[Key, Value]

logger = logging.getLogger(__name__)


class KeyValueDb(Generic[Key, Value]):
    """
    Key-value in-memory database
    Use open() method to load DB from a file and commit() to save it
    """

    _name = None  # type: str
    _records = None  # type: Records

    def __init__(self, basename: str):
        """
        :param str basename: path to db file
        """
        self._name = basename
        self.set_records({})

    def open(self) -> 'KeyValueDb':
        """Create a new data base or open existing one"""
        if os.path.exists(self._name):
            if not os.path.isfile(self._name):
                raise IOError("%s exists and is not a file" % self._name)

            if os.path.getsize(self._name) > 0:
                with open(self._name, 'rb') as reader:
                    try:
                        records = pickle.load(reader)
                        self.set_records(records)
                        return self
                    except (pickle.UnpicklingError, UnicodeDecodeError):
                        logger.error("Corrupted pickle file: %s. Will overwrite.", self._name)

        # make sure path exists
        os.makedirs(os.path.dirname(self._name), exist_ok=True)
        self.commit()
        return self

    def commit(self) -> 'KeyValueDb':
        """Write the database to a file"""
        with open(self._name, 'wb') as out:
            pickle_protocol = 2  # use 2 for BC with Ulauncher 4
            pickle.dump(self._records, out, pickle_protocol)
            out.close()

        return self

    def remove(self, key: Key) -> bool:
        """
        :param str key:
        :type: bool
        :return: True if record was removed
        """
        try:
            del self._records[key]
            return True
        except KeyError:
            return False

    def set_records(self, records: Records):
        self._records = records

    def get_records(self) -> Records:
        return self._records

    def put(self, key: Key, value: Value) -> None:
        self._records[key] = value

    def find(self, key: Key, default: Value = None) -> Optional[Value]:
        return self._records.get(key, default)



from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.api.shared.action.SetUserQueryAction import SetUserQueryAction
from ulauncher.search.QueryHistoryDb import QueryHistoryDb


class ExtensionKeywordResultItem(ResultItem):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._query_history = QueryHistoryDb.get_instance()

    def selected_by_default(self, query):
        """
        :param ~ulauncher.search.Query.Query query:
        """
        return self._query_history.find(query) == self.get_name()

    def on_enter(self, query):
        """
        :param ~ulauncher.search.Query.Query query: query
        """
        self._query_history.save_query(query, self.get_name())
        return SetUserQueryAction('%s ' % self.get_keyword())



from collections import namedtuple, OrderedDict


def namedtuple_from_dict(obj):
    if isinstance(obj, dict):
        fields = sorted(obj.keys())
        namedtuple_type = namedtuple(
            typename='GenericObject',
            field_names=fields,
            rename=True,
        )
        field_value_pairs = OrderedDict(
            (str(field), namedtuple_from_dict(obj[field]))
            for field in fields
        )
        try:
            return namedtuple_type(**field_value_pairs)
        except TypeError:
            # Cannot create namedtuple instance so fallback to dict (invalid attribute names)
            return dict(**field_value_pairs)
    elif isinstance(obj, (list, set, tuple, frozenset)):
        return [namedtuple_from_dict(item) for item in obj]
    else:
        return obj



import logging
import gi
gi.require_version('Gdk', '3.0')
gi.require_version('GdkX11', '3.0')
# pylint: disable=wrong-import-position

from gi.repository import Gdk, GdkX11  # type: ignore

logger = logging.getLogger(__name__)


def get_current_screen(window=None):
    """
    :rtype int:
    """
    default_screen = Gdk.Screen.get_default()
    try:
        if window:
            screen = default_screen.get_monitor_at_window(window.get_window())
        else:
            disp = GdkX11.X11Display.get_default()
            dm = Gdk.Display.get_device_manager(disp)
            pntr_device = dm.get_client_pointer()
            (_, x, y) = pntr_device.get_position()
            screen = default_screen.get_monitor_at_point(x, y)
    # pylint: disable=broad-except
    except Exception as e:
        logger.exception("Unexpected exception: %s", e)
        screen = 0

    return screen


def get_primary_screen_geometry():
    """
    :returns: dict with keys: x, y, width, height
    """
    return get_screens()[Gdk.Screen.get_default().get_primary_monitor()]


def get_monitor_scale_factor() -> int:
    display = Gdk.Display.get_default()
    if display:
        monitor = display.get_primary_monitor() or display.get_monitor(0)
        if monitor:
            return monitor.get_scale_factor()

    return 1


def get_current_screen_geometry(window=None):
    """
    :returns: dict with keys: x, y, width, height
    """
    return get_screens()[get_current_screen(window)]


def get_screens():
    """
    :returns: a list of screen geometries
    :raises RuntimeError:
    """

    screens = []
    try:
        default_screen = Gdk.Screen.get_default()
        logger.debug("Found %s monitor(s)", default_screen.get_n_monitors())

        for i in range(default_screen.get_n_monitors()):
            rect = default_screen.get_monitor_geometry(i)
            logger.debug("  Monitor %s - X: %s, Y: %s, W: %s, H: %s", i, rect.x, rect.y, rect.width, rect.height)
            screens.append({"x": rect.x,
                            "y": rect.y,
                            "width": rect.width,
                            "height": rect.height})
    except Exception as e:
        raise RuntimeError("Unable to find any video sources: %s" % e) from e

    return screens



import logging
import gi
gi.require_version('GObject', '2.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
# pylint: disable=wrong-import-position
from gi.repository import Gtk, Gdk, GObject

from ulauncher.ui.windows.Builder import Builder
from ulauncher.ui.windows.WindowHelper import WindowHelper


logger = logging.getLogger('ulauncher')

FORBIDDEN_ACCEL_KEYS = ('Delete', 'Page_Down', 'Page_Up', 'Home', 'End', 'Up', 'Down', 'Left', 'Right', 'Return',
                        'BackSpace', 'Alt_L', 'Alt_R', 'Shift_L', 'Shift_R', 'Control_L', 'Control_R', 'space',
                        'Escape', 'Tab', 'Insert')


class HotkeyDialog(Gtk.Dialog, WindowHelper):
    __gtype_name__ = "HotkeyDialog"

    __gsignals__ = {
        # parameters: <hotkey-value (str)>, <hotkey-display-value (str)>
        'hotkey-set': (GObject.SignalFlags.RUN_LAST, GObject.TYPE_NONE, (GObject.TYPE_STRING, GObject.TYPE_STRING))
    }

    _accel_name = None
    _display_name = None

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated HotkeyDialog object.
        """
        builder = Builder.new_from_file('HotkeyDialog')
        new_object = builder.get_object("hotkey_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initializing should be called after parsing the ui definition
        and creating a HotkeyDialog object with it in order
        to finish initializing the start of the new HotkeyDialog
        instance.

        Put your initialization code in here and leave __init__ undefined.
        """
        # pylint: disable=attribute-defined-outside-init
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)
        self.ui['hotkey_dialog_action_area'].destroy()

    def on_delete_event(self, *args):
        # don't delete. Hide instead
        self.hide()
        return True

    def on_hotkey_input_key_press_event(self, _, event):
        # remove GDK_MOD2_MASK, because it seems unnecessary
        mask = event.state
        if mask & Gdk.ModifierType.MOD2_MASK:
            mask ^= Gdk.ModifierType.MOD2_MASK
        if mask & Gdk.ModifierType.MOD4_MASK:
            mask ^= Gdk.ModifierType.MOD4_MASK

        accel_name = Gtk.accelerator_name(event.keyval, mask)
        display_name = Gtk.accelerator_get_label(event.keyval, mask)

        if accel_name == 'Return':
            # emit hotkey-set signal
            self.emit('hotkey-set', self._accel_name, self._display_name)
            self.hide()

        if accel_name == 'Escape':
            self.hide()
            return

        # do nothing for invalid hotkeys
        if not self.is_valid_hotkey(display_name, accel_name):
            logger.debug("Invalid hotkey '%s', ('%s') is not allowed", display_name, accel_name)
            return

        self._accel_name = accel_name
        self._display_name = display_name

        self.ui['hotkey_input'].set_text(display_name)

    def is_valid_hotkey(self, label, accel_name):
        """
        :param str label: String like 'Ctrl+`'
        :param str accel_name: String like '<Primary>space'

        accel_name should not be in FORBIDDEN_ACCEL_KEYS
        and label should consist of more than one symbol

        NOTE: it's very likely these verifications are not enough,
        but I could not find more information about this topic
        """
        return (accel_name not in FORBIDDEN_ACCEL_KEYS) and len(label) > 1 and not accel_name.startswith('KP_')



# pylint: disable-all
import gi

# The package version refers to the "build environment". Not the actual Webkit version
# 4.0 means it was built with GTK 3 and libsoup 2
# 4.1 means it was built with GTK 3 and libsoup 3 (HTTP/2)
# 5.0 means it was built with GTK 4 and libsoup 3 (incompatible w Ulauncher, and was only supported for a short period)
# 6.0 means the same as 5.0, but the API has major changes, and the package name was renamed, dropping the "2"
# https://blog.tingping.se/2021/06/07/http2-in-libsoup.html
# https://discourse.gnome.org/t/please-build-against-libsoup-3-by-default/10190
# https://bugs.webkit.org/show_bug.cgi?id=245296
try:
    gi.require_version('WebKit2', '4.1')
except ValueError:
    gi.require_version('WebKit2', '4.0')
from gi.repository import WebKit2  # type: ignore[attr-defined] # noqa: F401



import gi
try:
    gi.require_version("Notify", "0.7")
except ValueError:
    gi.require_version("Notify", "0.8")

# pylint: disable=wrong-import-position
from gi.repository import Notify  # type: ignore

Notify.init('ulauncher')


def show_notification(summary, body, icon='ulauncher'):
    """
    :rtype: :class:`Notify.Notification`
    """
    Notify.Notification.new(summary, body, icon).show()



class BaseAction:

    def keep_app_open(self):
        """
        :rtype: bool
        :returns: True if Ulauncher window should remain open once all actions are done
        """
        return False

    def run(self):
        """
        Runs action
        """
        raise RuntimeError("%s#run() is not implemented" % self.__class__.__name__)



import os
import logging
import json
from typing import Dict, Any
from shutil import copytree, rmtree

from ulauncher.config import get_data_path, CONFIG_DIR, CACHE_DIR
from ulauncher.utils.Settings import Settings
from ulauncher.utils.version_cmp import gtk_version_is_gte

themes: Dict[str, Any] = {}
logger = logging.getLogger(__name__)
user_theme_dir = os.path.join(CONFIG_DIR, 'user-themes')


def load_available_themes():
    themes.clear()
    ulauncher_theme_dir = os.path.join(get_data_path(), 'themes')
    theme_dirs = [os.path.join(ulauncher_theme_dir, d) for d in os.listdir(ulauncher_theme_dir)]
    if os.path.exists(user_theme_dir):
        theme_dirs.extend([os.path.join(user_theme_dir, d) for d in os.listdir(user_theme_dir)])
    theme_dirs = list(filter(os.path.isdir, theme_dirs))

    for dir in theme_dirs:
        if os.path.isfile(os.path.join(dir, 'manifest.json')):
            theme = Theme(dir)
            try:
                theme.validate()
            except ThemeManifestError as e:
                logger.error('%s: %s', type(e).__name__, e)
                continue
            themes[theme.get('name')] = theme


class Theme:

    @classmethod
    def get_current(cls):
        default = 'light'
        current_name = Settings.get_instance().get_property('theme-name') or default
        try:
            current = themes[current_name]
        except KeyError:
            logger.warning('No theme with name %s', current_name)
            current = themes[default]

        return current

    _data = None

    def __init__(self, path):
        self.path = path

    def clear_cache(self):
        self._data = None

    def _read(self):
        if not self._data:
            with open(os.path.join(self.path, 'manifest.json'), 'r') as file:
                self._data = json.load(file)

        return self._data

    def get(self, property: str):
        return self._read().get(property)

    def validate(self):
        try:
            assert self.get('manifest_version') in ['1'], "Supported manifest version is '1'"
            for prop in ['name', 'display_name', 'matched_text_hl_colors', 'css_file', 'css_file_gtk_3.20+']:
                assert self.get(prop), '"%s" is empty' % prop

            assert os.path.exists(os.path.join(self.path, self.get('css_file'))), '"css_file" does not exist'
            assert os.path.exists(os.path.join(self.path, self.get('css_file_gtk_3.20+'))), \
                '"css_file_gtk_3.20+" does not exist'
        except AssertionError as e:
            raise ThemeManifestError(e) from e

    def compile_css(self) -> None:
        # workaround for issue with a caret-color
        # GTK+ < 3.20 doesn't support that prop
        css_file_name = self.get('css_file_gtk_3.20+') if gtk_version_is_gte(3, 20, 0) else self.get('css_file')
        css_file = os.path.join(self.path, css_file_name)

        if not self.get('extend_theme'):
            return css_file

        # if theme extends another one, we must import it in css
        # therefore a new css file (generated.css) is created here
        extend_theme_name = self.get('extend_theme')
        try:
            extend_theme = themes[extend_theme_name]
        except KeyError:
            logger.error('Cannot extend theme "%s". It does not exist', extend_theme_name)
            return css_file

        generated_css = self._get_path_for_generated_css()
        with open(generated_css, 'w') as new_css_file:
            new_css_file.write('@import url("%s");\n\n' % extend_theme.compile_css())
            with open(css_file, 'r') as theme_css_file:
                new_css_file.write(theme_css_file.read())

        return generated_css

    def _get_path_for_generated_css(self):
        if user_theme_dir in self.path:
            return os.path.join(self.path, 'generated.css')

        # for ulauncher themes we must save generated.css elsewhere
        # because we don't have write permissions for /usr/share/ulauncher/themes/...
        new_theme_dir = os.path.join(CACHE_DIR, 'themes', self.get('name'))
        if not os.path.exists(new_theme_dir):
            os.makedirs(new_theme_dir)

        # copy current theme files to this new dir
        rmtree(new_theme_dir)
        copytree(self.path, new_theme_dir)

        # ensure correct file permissions
        os.chmod(new_theme_dir, 0o755)

        return os.path.join(new_theme_dir, 'generated.css')


class ThemeManifestError(Exception):
    pass



import pickle

from ulauncher.api.shared.event import ItemEnterEvent
from ulauncher.api.shared.action.BaseAction import BaseAction


class ExtensionCustomAction(BaseAction):
    """
    If initiated with ``data``, the same data will be returned
    in :class:`~ulauncher.api.shared.event.ItemEnterEvent` object

    :param any data: any type that can be serialized with :func:`pickle.dumps`
    :param bool keep_app_open: pass ``True`` if you want to keep Ulauncher window open.
                               ``False`` by default
    """

    def __init__(self, data, keep_app_open=False):
        self._data = pickle.dumps(data)
        self._keep_app_open = keep_app_open

    def keep_app_open(self):
        return self._keep_app_open

    def run(self):
        """
        Runs :meth:`~ulauncher.api.server.ExtensionController.ExtensionController.trigger_event`
        with :class:`ItemEnterEvent`
        """
        # import here to avoid circular deps
        from ulauncher.api.server.DeferredResultRenderer import DeferredResultRenderer
        from ulauncher.ui.windows.UlauncherWindow import UlauncherWindow

        window = UlauncherWindow.get_instance()
        if not window.is_visible():
            return

        renderer = DeferredResultRenderer.get_instance()
        controller = renderer.get_active_controller()
        if controller:
            controller.trigger_event(ItemEnterEvent(self._data))



import re
from urllib.parse import unquote

RE_URL = re.compile(r'^(?P<scheme>.*)://(?P<path>[^\?]*)(\?(?P<query>.*))?$', flags=re.IGNORECASE)


def get_url_params(url):
    params = re.search(RE_URL, url)
    query = params.group('query')
    if query:
        pairs = list(map(lambda kv: kv.split('='), query.split('&')))
        query = {k: unquote(v) for k, v in pairs}
    return {
        'scheme': params.group('scheme'),
        'path': params.group('path'),
        'query': query or None
    }


class Router:
    """
    Usage:

    >>> rt = Router()
    >>>
    >>> class App:
    >>>
    >>>     @rt.route('get/user')
    >>>     def get_user(self, url_params):
    >>>         ...
    >>>         return userObject
    >>>
    >>>     def on_request(self, url):
    >>>         result = rt.dispatch(self, url) # will return userObject for /get/user path
    """

    _callbacks = None

    def __init__(self):
        self._callbacks = {}

    def dispatch(self, context, url):
        url_params = get_url_params(url)
        try:
            callback = self._callbacks[url_params['path'].strip('/')]
        except KeyError as e:
            raise RouteNotFound('Route not found for path %s' % url_params['path']) from e

        return callback(context, url_params)

    def route(self, path):
        if not path:
            raise RoutePathEmpty()

        def decorator(callback):
            self._callbacks[path.strip('/')] = callback
            return callback

        return decorator


class RouterError(RuntimeError):
    pass


class RoutePathEmpty(RouterError):
    pass


class RouteNotFound(RouterError):
    pass



import os
import sys

from ulauncher.api.shared.action.BaseAction import BaseAction
from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.utils.image_loader import load_image
from ulauncher.search.QueryHistoryDb import QueryHistoryDb


class ExtensionResultItem(ResultItem):
    """
    Should be used in extensions.

    Cannot be subclassed there because :func:`pickle.loads` won't work in Ulauncher app
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._is_extension = True
        self.extension_path = os.path.dirname(sys.argv[0])
        if self._on_enter and not isinstance(self._on_enter, BaseAction):
            raise Exception("Incorrect type of on_enter argument")

    def get_icon(self):
        if isinstance(self._icon, str):
            icon_path = self._icon

            if not icon_path.startswith('/'):
                icon_path = os.path.join(self.extension_path, icon_path)

            return load_image(icon_path, self.get_icon_size())

        # assuming it's GtkPixbuf
        return self._icon

    def on_enter(self, query):
        return self._on_enter

    def on_alt_enter(self, query):
        return self._on_alt_enter

    def selected_by_default(self, query):
        query_history = QueryHistoryDb.get_instance()
        return query_history.find(query) == self.get_name()



from ulauncher.api.shared.action.LaunchAppAction import LaunchAppAction
from ulauncher.api.shared.item.ResultItem import ResultItem
from ulauncher.search.QueryHistoryDb import QueryHistoryDb
from ulauncher.search.apps.AppStatDb import AppStatDb


class AppResultItem(ResultItem):
    """
    :param dict record:
    """

    # pylint: disable=super-init-not-called
    def __init__(self, record):
        self.record = record
        self._query_history = QueryHistoryDb.get_instance()
        self._app_stat_db = AppStatDb.get_instance()

    def get_name(self):
        return self.record.get('name')

    def get_search_name(self):
        return self.record.get('search_name')

    def get_description(self, query):
        return self.record.get('description')

    def selected_by_default(self, query):
        """
        :param ~ulauncher.search.Query.Query query:
        """
        return self._query_history.find(query) == self.record.get('name')

    def get_icon(self):
        return self.record.get('icon')

    def on_enter(self, query):
        self._query_history.save_query(str(query), self.record.get('name'))

        self._app_stat_db.inc_count(self.record.get('desktop_file'))
        self._app_stat_db.commit()

        return LaunchAppAction(self.record.get('desktop_file'))



from ulauncher.api.shared.action.BaseAction import BaseAction


class DoNothingAction(BaseAction):
    """
    Does nothing. Keeps Ulauncher window open
    """

    def keep_app_open(self):
        return True

    def run(self):
        pass



# pylint: disable=too-few-public-methods
class EventListener:
    """
    Base event listener class
    """

    def on_event(self, event, extension):
        """
        :param ~ulauncher.api.shared.event.BaseEvent event: event that listener was subscribed to
        :param ~ulauncher.api.client.Extension.Extension extension:

        :rtype: :class:`~ulauncher.api.shared.action.BaseAction.BaseAction` or None
        :return: Instance of :class:`BaseAction` or :code:`None`
        """



import logging

from ulauncher.api.server.ExtensionSearchMode import ExtensionSearchMode
from ulauncher.search.apps.AppSearchMode import AppSearchMode
from ulauncher.search.shortcuts.ShortcutSearchMode import ShortcutSearchMode
from ulauncher.search.file_browser.FileBrowserMode import FileBrowserMode
from ulauncher.search.calc.CalcMode import CalcMode
from ulauncher.utils.decorator.singleton import singleton

logger = logging.getLogger(__name__)


class Search:

    @classmethod
    @singleton
    def get_instance(cls):
        file_browser_mode = FileBrowserMode()
        calc_mode = CalcMode()
        shortcut_search_mode = ShortcutSearchMode()
        extension_search_mode = ExtensionSearchMode()
        app_search_mode = AppSearchMode([shortcut_search_mode, extension_search_mode])
        return cls([file_browser_mode,
                    calc_mode,
                    shortcut_search_mode,
                    extension_search_mode,
                    app_search_mode])

    def __init__(self, search_modes):
        self.search_modes = search_modes

    def on_query_change(self, query):
        """
        Iterate over all search modes and run first enabled.
        AppSearchMode is always enabled
        """
        for mode in self.search_modes:
            mode.on_query_change(query)

        self._choose_search_mode(query).handle_query(query).run()

    def on_key_press_event(self, widget, event, query):
        self._choose_search_mode(query).handle_key_press_event(widget, event, query).run()

    def _choose_search_mode(self, query):
        for mode in self.search_modes:
            if mode.is_enabled(query):
                return mode

        raise Exception('This line should not be entered')



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

import os
import re
import sys

try:
    import DistUtilsExtra.auto
except ImportError:
    print('To build ulauncher you need "python3-distutils-extra"', file=sys.stderr)
    sys.exit(1)

assert DistUtilsExtra.auto.__version__ >= '2.18', \
    'needs DistUtilsExtra.auto >= 2.18'


def update_config(libdir, values=None):
    if not values:
        values = {}
    filename = os.path.join(libdir, 'ulauncher/config.py')
    oldvalues = {}
    try:
        fin = open(filename, 'r')
        fout = open(filename + '.new', 'w')

        for line in fin:
            fields = line.split(' = ')  # Separate variable from value
            if fields[0] in values:
                oldvalues[fields[0]] = fields[1].strip()
                line = "%s = %s\n" % (fields[0], values[fields[0]])
            fout.write(line)

        fout.flush()
        fout.close()
        fin.close()
        os.rename(fout.name, fin.name)
    except OSError:
        print("ERROR: Can't find %s" % filename)
        sys.exit(1)

    return oldvalues


def move_desktop_file(root, target_data, prefix):
    # The desktop file is rightly installed into install_data.  But it should
    # always really be installed into prefix, because while we can install
    # normal data files anywhere we want, the desktop file needs to exist in
    # the main system to be found.  Only actually useful for /opt installs.

    old_desktop_path = os.path.normpath(root + target_data + '/share/applications')
    old_desktop_file = old_desktop_path + '/ulauncher.desktop'
    desktop_path = os.path.normpath(root + prefix + '/share/applications')
    desktop_file = desktop_path + '/ulauncher.desktop'

    if not os.path.exists(old_desktop_file):
        print("ERROR: Can't find", old_desktop_file)
        sys.exit(1)
    elif os.path.normpath(target_data) != os.path.normpath(prefix):
        # This is an /opt install, so rename desktop file to use extras-
        desktop_file = desktop_path + '/extras-ulauncher.desktop'
        try:
            os.makedirs(desktop_path)
            os.rename(old_desktop_file, desktop_file)
            os.rmdir(old_desktop_path)
        except OSError as e:
            print("ERROR: Can't rename", old_desktop_file, ":", e)
            sys.exit(1)

    return desktop_file


def update_desktop_file(filename, target_pkgdata, target_scripts):
    try:
        with open(filename, "r") as fin:
            src = fin.read()

        dst = re.sub(
            r"((Try)?Exec)=(.*?)(/[^ ]+/)?ulauncher(.*)",
            r"\1=\3{}ulauncher\5".format(target_scripts),
            src
        )

        with open(filename, "w") as fout:
            fout.write(dst)
    except OSError:
        print("ERROR: Can't find %s" % filename)
        sys.exit(1)


class InstallAndUpdateDataDirectory(DistUtilsExtra.auto.install_auto):
    def run(self):
        DistUtilsExtra.auto.install_auto.run(self)

        # Root is undefined if not installing into an alternate root
        root = self.root or "/"
        target_data = '/' + os.path.relpath(self.install_data, root) + '/'
        target_pkgdata = target_data + 'share/ulauncher/'
        target_scripts = '/' + os.path.relpath(self.install_scripts,
                                               root) + '/'

        values = {'__ulauncher_data_directory__': "'%s'" % (target_pkgdata),
                  '__version__': "'%s'" % self.distribution.get_version()}
        update_config(self.install_lib, values)

        desktop_file = move_desktop_file(root, target_data, self.prefix)
        update_desktop_file(desktop_file, target_pkgdata, target_scripts)


class DataFileList(list):

    def append(self, item):
        # don't add node_modules to data_files that DistUtilsExtra tries to
        # add automatically
        filename = item[1][0]
        if 'node_modules' in filename \
           or 'bower_components' in filename or '.tmp' in filename:
            return
        else:
            return super().append(item)


def exclude_files(patterns=None):
    """
    Suppress completely useless warning about files DistUtilsExta.aut does
    recognize because require developer to scroll past them to get to useful
    output.

    Example of the useless warnings:

    WARNING: the following files are not recognized by DistUtilsExtra.auto:
    Dockerfile
    PKGBUILD.template
    """

    # it's maddening the DistUtilsExtra does not offer a way to exclude globs
    # from it's scans and just using "print" to print the warning instead of
    # using warning module which has a mechanism for suppressions
    # it forces us to take the approach of monkeypatching their src_find
    # function.
    original_src_find = DistUtilsExtra.auto.src_find

    def src_find_with_excludes(attrs):
        src = original_src_find(attrs)

        for pattern in patterns or []:
            DistUtilsExtra.auto.src_markglob(src, pattern)

        return src

    DistUtilsExtra.auto.src_find = src_find_with_excludes

    return original_src_find


def main():

    # exclude files/folder patterns from being considered by distutils-extra
    # this returns the original DistUtilsExtra.auto.src_find function
    # so we can patch bit back in later
    original_find_src = exclude_files([
        "*.sh",
        "ul",
        "Dockerfile.build*",
        "PKGBUILD.template",
        "scripts/*",
        "docs/*",
        "glade",
        "test",
        "ulauncher.desktop.dev",
        "requirements.txt",
        "conftest.py"
    ])

    DistUtilsExtra.auto.setup(
        name='ulauncher',
        version='5.15.15',
        license='GPL-3',
        author='Aleksandr Gornostal',
        author_email='ulauncher.app@gmail.com',
        description='Application launcher for Linux',
        url='https://ulauncher.io',
        data_files=DataFileList([
            ('share/icons/hicolor/48x48/apps', [
                'data/media/icons/hicolor/ulauncher.svg'
            ]),
            ('share/icons/hicolor/48x48/apps', [
                'data/media/icons/hicolor/ulauncher-indicator.svg'
            ]),
            ('share/icons/hicolor/scalable/apps', [
                'data/media/icons/hicolor/ulauncher.svg'
            ]),
            ('share/icons/hicolor/scalable/apps', [
                'data/media/icons/hicolor/ulauncher-indicator.svg'
            ]),
            # for fedora + GNOME
            ('share/icons/gnome/scalable/apps', [
                'data/media/icons/hicolor/ulauncher.svg'
            ]),
            ('share/icons/gnome/scalable/apps', [
                'data/media/icons/hicolor/ulauncher-indicator.svg'
            ]),
            # for ubuntu
            ('share/icons/breeze/apps/48', [
                'data/media/icons/ubuntu-mono-light/ulauncher-indicator.svg'
            ]),
            ('share/icons/ubuntu-mono-dark/scalable/apps', [
                'data/media/icons/hicolor/ulauncher-indicator.svg'
            ]),
            ('share/icons/ubuntu-mono-light/scalable/apps', [
                'data/media/icons/ubuntu-mono-light/ulauncher-indicator.svg'
            ]),
            ('share/icons/elementary/scalable/apps', [
                'data/media/icons/elementary/ulauncher-indicator.svg'
            ]),
            ('share/applications', [
                'build/share/applications/ulauncher.desktop'
            ]),
            ('lib/systemd/user', [
                'ulauncher.service'
            ])
        ]),
        cmdclass={'install': InstallAndUpdateDataDirectory}
    )

    # unpatch distutils-extra src_find
    DistUtilsExtra.auto.src_find = original_find_src


if __name__ == '__main__':
    main()



import logging
import json
import os
from pathlib import Path
from ulauncher.utils.semver import satisfies
from ulauncher.api.version import api_version

logger = logging.getLogger(__name__)


def find_extensions(ext_dir):
    """
    Yields `(extension_id, extension_path)` tuples found in a given extensions dir
    """
    if not os.path.exists(ext_dir):
        return

    dirs = [d for d in os.listdir(ext_dir) if os.path.isdir(os.path.join(ext_dir, d))]
    for dir in dirs:
        ext_path = os.path.join(ext_dir, dir)
        if os.path.isfile(os.path.join(ext_path, 'manifest.json')):
            try:
                manifest = json.loads(Path(ext_path, 'manifest.json').read_text('utf-8'))
                api_version_range = manifest.get('required_api_version', manifest.get('api_version'))
                if satisfies(api_version, api_version_range):
                    yield (dir, ext_path)
                else:
                    logger.warning(
                        'Ignoring incompatible extension %s (extension API expected %s, but the current version is %s)',
                        dir, api_version_range, api_version
                    )
            except Exception as e:
                logger.error('Ignoring extension %s (has a broken manifest.json file): %s', dir, e)



"""
mypy_extensions cannot be installed via apt or dnf so we need to make sure TypedDict exists at runtime
"""

# pylint: disable=unused-import
from typing import Any, Callable, Dict  # noqa: F401


def _TypedDict(*args, **kw):
    pass


try:
    from mypy_extensions import TypedDict
except ImportError:
    TypedDict = _TypedDict



from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.search.BaseSearchMode import BaseSearchMode
from ulauncher.search.apps.AppDb import AppDb


class AppSearchMode(BaseSearchMode):
    """
    :type list search_modes: a list of other :class:`SearchMode` objects that provide additional result items
    """

    def __init__(self, search_modes):
        self.search_modes = search_modes
        self.app_db = AppDb.get_instance()

    def is_enabled(self, query):
        """
        AppSearchMode is a default search mode and is always enabled
        """
        return True

    def handle_query(self, query):
        result_list = self.app_db.find(query)
        for mode in self.search_modes:
            result_list.extend(mode.get_searchable_items())

        if not result_list and query:
            # default search
            result_list = []
            for mode in self.search_modes:
                result_list.extend(mode.get_default_items())

        return RenderResultListAction(result_list)



from enum import Enum


class ErrorName(Enum):
    InvalidGithubUrl = 'InvalidGithubUrl'
    IncompatibleVersion = 'IncompatibleVersion'
    InvalidVersionsJson = 'InvalidVersionsJson'
    GithubApiError = 'GithubApiError'
    ExtensionAlreadyAdded = 'ExtensionAlreadyAdded'
    UnexpectedError = 'UnexpectedError'
    UnhandledError = 'UnhandledError'
    InvalidManifestJson = 'InvalidManifestJson'
    ExtensionCompatibilityError = 'ExtensionCompatibilityError'


class UlauncherAPIError(Exception):
    error_name = None  # type: str

    def __init__(self, message: str, error_name: ErrorName = ErrorName.UnexpectedError):
        super().__init__(message)
        self.error_name = error_name.value



import os
import json
from typing import cast, Optional, List, Union
from ulauncher.config import EXTENSIONS_DIR
from ulauncher.utils.image_loader import load_image
from ulauncher.api.shared.errors import UlauncherAPIError, ErrorName
from ulauncher.api.version import api_version
from ulauncher.utils.semver import satisfies
from ulauncher.utils.mypy_extensions import TypedDict


class ExtensionManifestError(UlauncherAPIError):
    pass


OptionItemExtended = TypedDict('OptionItemExtended', {
    'value': str,
    'text': str
})
OptionItem = Union[str, OptionItemExtended]
OptionItems = List[OptionItem]
Options = TypedDict('Options', {
    'query_debounce': float
})
ManifestPreferenceItem = TypedDict('ManifestPreferenceItem', {
    'id': str,
    'type': str,
    'name': str,
    'description': str,
    'default_value': str,
    'options': OptionItems
})
ManifestJson = TypedDict('ManifestJson', {
    'required_api_version': str,
    'name': str,
    'description': str,
    'developer_name': str,
    'icon': str,
    'options': Optional[Options],
    'preferences': List[ManifestPreferenceItem]
})


class ExtensionManifest:
    """
    Reads `manifest.json`
    """
    manifest = None  # type: ManifestJson

    @classmethod
    def open(cls, extension_id, extensions_dir=EXTENSIONS_DIR):
        return cls(extension_id, read_manifest(extension_id, extensions_dir), extensions_dir)

    def __init__(self, extension_id: str, manifest: ManifestJson, extensions_dir: str = EXTENSIONS_DIR):
        self.extensions_dir = extensions_dir
        self.extension_id = extension_id
        self.manifest = manifest

    def refresh(self):
        self.manifest = cast(ManifestJson, read_manifest(self.extension_id, self.extensions_dir))

    def get_name(self) -> str:
        return self.manifest['name']

    def get_description(self) -> str:
        return self.manifest['description']

    def get_icon(self) -> str:
        return self.manifest['icon']

    def get_icon_path(self) -> str:
        return os.path.join(self.extensions_dir, self.extension_id, self.get_icon())

    def load_icon(self, size):
        return load_image(self.get_icon_path(), size)

    def get_required_api_version(self) -> str:
        return self.manifest['required_api_version']

    def get_developer_name(self) -> str:
        return self.manifest['developer_name']

    def get_preferences(self) -> List[ManifestPreferenceItem]:
        return self.manifest.get('preferences', [])

    def get_preference(self, id) -> Optional[ManifestPreferenceItem]:
        for p in self.get_preferences():
            if p['id'] == id:
                return p

        return None

    def get_option(self, name, default=None):
        try:
            return self.manifest['options'][name]
        except KeyError:
            return default

    def validate(self):
        try:
            assert self.get_required_api_version(), "required_api_version is not provided"
            assert self.get_name(), 'name is not provided'
            assert self.get_description(), 'description is not provided'
            assert self.get_developer_name(), 'developer_name is not provided'
            assert self.get_icon(), 'icon is not provided'

            for p in self.get_preferences():
                assert p.get('id'), 'Preferences error. Id cannot be empty'
                assert p.get('type'), 'Preferences error. Type cannot be empty'
                assert p.get('type') in ["keyword", "input", "text", "select"], \
                    'Preferences error. Type can be "keyword", "input", "text", or "select"'
                assert p.get('name'), 'Preferences error. Name cannot be empty'
                if p['type'] == 'keyword':
                    assert p.get('default_value'), 'Preferences error. Default value cannot be empty for keyword'
                if p['type'] == 'select':
                    assert isinstance(p.get('options'), list), 'Preferences error. Options must be a list'
                    assert p.get('options'), 'Preferences error. Option list cannot be empty'
        except AssertionError as e:
            raise ExtensionManifestError(str(e), ErrorName.InvalidManifestJson) from e
        except KeyError as e:
            raise ExtensionManifestError('%s is not provided' % e, ErrorName.InvalidManifestJson) from e

    def check_compatibility(self):
        if not satisfies(api_version, self.get_required_api_version()):
            raise ExtensionManifestError('Extension "%s" requires Ulauncher API %s, but current API version is %s' %
                                         (self.extension_id, self.get_required_api_version(), api_version),
                                         ErrorName.ExtensionCompatibilityError)


def read_manifest(extension_id, extensions_dir):
    with open(os.path.join(extensions_dir, extension_id, 'manifest.json'), 'r') as f:
        return json.load(f)



from gi.repository import Gio

try:
    SystemDesktopAppInfo = Gio.DesktopAppInfo
except AttributeError:
    import gi

    try:
        gi.require_version("GioUnix", "2.0")
        from gi.repository import GioUnix  # type: ignore[attr-defined]

        SystemDesktopAppInfo = GioUnix.DesktopAppInfo  # type: ignore[assignment,misc,unused-ignore]
    except (ImportError, ValueError, AttributeError):
        raise ImportError("Could not import Gio.DesktopAppInfo")


class DesktopAppInfo:
    _app_info: Gio.DesktopAppInfo
    """
    Wrapper for Gio.DesktopAppInfo because it is broken in Debian Forky/Testing
    """

    def __init__(self, app_info: Gio.DesktopAppInfo) -> None:
        self._app_info = app_info

    @staticmethod
    def new(app_id):
        app_info = SystemDesktopAppInfo.new(app_id)
        if app_info:
            return DesktopAppInfo(app_info)
        return None

    @staticmethod
    def new_from_filename(filename):
        app_info = SystemDesktopAppInfo.new_from_filename(filename)
        if app_info:
            return DesktopAppInfo(app_info)
        return None

    @staticmethod
    def get_all():
        return [DesktopAppInfo(app_info) for app_info in SystemDesktopAppInfo.get_all()]  # type: ignore[arg-type]

    # Methods copied from Gio.DesktopAppInfo (these work fine)

    def get_id(self):
        return self._app_info.get_id()

    def get_name(self):
        return self._app_info.get_name()

    def get_description(self):
        return self._app_info.get_description()

    def get_display_name(self):
        return self._app_info.get_display_name()

    def get_commandline(self):
        return self._app_info.get_commandline()

    def get_executable(self):
        return self._app_info.get_executable()

    def get_icon(self):
        return self._app_info.get_icon()

    # Borked unbound methods that we have to call via the class now to work consistently
    def get_boolean(self, name):
        return SystemDesktopAppInfo.get_boolean(self._app_info, name)

    def get_string(self, name):
        return SystemDesktopAppInfo.get_string(self._app_info, name)

    def get_generic_name(self):
        return SystemDesktopAppInfo.get_generic_name(self._app_info)

    def get_filename(self):
        return SystemDesktopAppInfo.get_filename(self._app_info)

    def get_keywords(self):
        return SystemDesktopAppInfo.get_keywords(self._app_info)

    def get_is_hidden(self):
        return SystemDesktopAppInfo.get_is_hidden(self._app_info)

    def get_show_in(self):
        return SystemDesktopAppInfo.get_show_in(self._app_info)

    def get_nodisplay(self):
        return SystemDesktopAppInfo.get_nodisplay(self._app_info)



import subprocess
from ulauncher.api.shared.action.BaseAction import BaseAction


class OpenAction(BaseAction):
    """
    Run platform specific command to open either file or directory

    :param str path: file or dir path
    """

    def __init__(self, path):
        self.path = path

    def keep_app_open(self):
        return False

    def run(self):
        subprocess.Popen(['xdg-open', self.path])



import os
import re
import sqlite3
import logging

from ulauncher.search.SortedList import SortedList
from ulauncher.utils.decorator.singleton import singleton
from ulauncher.search.apps.AppResultItem import AppResultItem
from ulauncher.search.apps.AppIconCache import AppIconCache

logger = logging.getLogger(__name__)


class AppDb:

    @classmethod
    @singleton
    def get_instance(cls):
        return cls(':memory:', AppIconCache.get_instance()).open()  # in-memory SqLite DB

    def __init__(self, path: str, app_icon_cache: AppIconCache):
        self._path = path
        self._app_icon_cache = app_icon_cache
        self._conn = None

    def open(self):
        create_db_scheme = self._path == ':memory:' or not os.path.exists(self._path)
        self._conn = sqlite3.connect(self._path, check_same_thread=False)
        self._conn.row_factory = sqlite3.Row
        if create_db_scheme:
            self._create_table()
        return self

    def get_cursor(self):
        return self._conn.cursor()

    # pylint: disable=unused-argument
    def commit(self, force=False):
        self._conn.commit()

    def _create_table(self):
        self._conn.executescript('''
            CREATE TABLE app_db (
              name VARCHAR,
              desktop_file VARCHAR,
              desktop_file_short VARCHAR,
              description VARCHAR,
              search_name VARCHAR,
              PRIMARY KEY (desktop_file_short)
            );
            CREATE INDEX desktop_file_idx ON app_db (desktop_file);
        ''')

    def _row_to_rec(self, row):
        """
        :param sqlite3.Row row:
        """
        return {
            'desktop_file': row['desktop_file'],
            'desktop_file_short': row['desktop_file_short'],
            'name': row['name'],
            'description': row['description'],
            'search_name': row['search_name'],
            'icon': self._app_icon_cache.get_pixbuf(row['desktop_file']),
        }

    def put_app(self, app):
        """
        :param Gio.DesktopAppInfo app:
        """
        name = app.get_string('X-GNOME-FullName') or app.get_name()
        exec_name = get_exec_name(app.get_string('Exec') or '')
        description = app.get_description() or ''
        if not description and (app.get_generic_name() != name):
            description = app.get_generic_name() or ''
        record = {
            "desktop_file": app.get_filename(),
            "desktop_file_short": os.path.basename(app.get_filename()),
            "description": app.get_description() or '',
            "name": name,
            "search_name": search_name(name, exec_name)
        }
        self._app_icon_cache.add_icon(record['desktop_file'], app.get_icon(), app.get_string('Icon'))

        query = '''INSERT OR REPLACE INTO app_db (name, desktop_file, desktop_file_short, description, search_name)
                   VALUES (:name, :desktop_file, :desktop_file_short, :description, :search_name)'''
        try:
            self._conn.execute(query, record)
            self.commit()
        # pylint: disable=broad-except
        except Exception as e:
            logger.exception('Exception %s for query: %s. Record: %s', e, query, record)

    def get_by_name(self, name):
        query = 'SELECT * FROM app_db where name = ? COLLATE NOCASE'
        try:
            collection = self._conn.execute(query, (name,))
        except Exception as e:
            logger.exception('Exception %s for query: %s. Name: %s', e, query, name)
            raise

        row = collection.fetchone()
        if row:
            return self._row_to_rec(row)

        return None

    def get_by_path(self, desktop_file):
        query = 'SELECT * FROM app_db where desktop_file = ?'
        try:
            collection = self._conn.execute(query, (desktop_file,))
        except Exception as e:
            logger.exception('Exception %s for query: %s. Path: %s', e, query, desktop_file)
            raise

        row = collection.fetchone()
        if row:
            return self._row_to_rec(row)

        return None

    def remove_by_path(self, desktop_file):
        """
        :param str desktop_file: path to a desktop file
        """
        query = 'DELETE FROM app_db WHERE desktop_file = ?'
        try:
            self._conn.execute(query, (desktop_file,))
            self.commit()
        except Exception as e:
            logger.exception('Exception %s for query: %s. Path: %s', e, query, desktop_file)
            raise

        self._app_icon_cache.remove_icon(desktop_file)

    def get_records(self):
        for row in self._conn.execute('SELECT * FROM app_db'):
            yield self._row_to_rec(row)

    def find(self, query, result_list=None):
        """
        :param str query: query to search for
        :param ResultList result_list:
        :rtype: :class:`ResultList`
        """

        result_list = result_list or SortedList(query, min_score=75, limit=9)

        if not query:
            return result_list

        for rec in self.get_records():
            result_list.append(AppResultItem(rec))

        return result_list


def get_exec_name(exec):
    """
    Returns name of the executable file without the full path, env keyword and environment vars
    Similar to Gio.DesktopAppInfo.get_executable(), except that will not drop env keyword and vars
    """
    match = re.match(r'^(env\s+)?([^\s=]+=[^\s=]+\s+)*([^\s=]+\/)*(?P<bin>[^\s=]*)', exec, re.I)
    return match.group('bin') if match else ""


def search_name(name, exec_name):
    """
    Returns string that will be used for search
    We want to make sure app can be searchable by its exec_line
    """
    return '{}\n{}'.format(name, exec_name)



import os
from operator import itemgetter
from itertools import islice

from ulauncher.config import DATA_DIR
from ulauncher.utils.db.KeyValueDb import KeyValueDb
from ulauncher.utils.decorator.singleton import singleton


class AppStatDb(KeyValueDb):
    """
    AppStatDb manages app launch statistics
    """

    @classmethod
    @singleton
    def get_instance(cls):
        return cls(os.path.join(DATA_DIR, 'app_stat_v2.db')).open()

    def inc_count(self, path):
        count = self._records.get(path, 0)
        self._records[path] = count + 1

    def get_most_frequent(self, limit=5):
        """
        Returns most frequent apps

        TODO: rename to `get_most_recent` and update method to remove old apps

        :param int limit: limit
        :rtype: class:`ResultList`
        """

        # pylint: disable=relative-beyond-top-level
        # import here to avoid circular deps.
        from ulauncher.search.apps.AppResultItem import AppResultItem
        from ulauncher.search.apps.AppDb import AppDb
        app_db = AppDb.get_instance()

        return [AppResultItem(i) for i in islice(
            filter(None,
                   map(lambda r: app_db.get_by_path(r[0]),
                       sorted(self._records.items(),
                              key=itemgetter(1),
                              reverse=True))),
            0, limit)]



import time
import os
import logging
import subprocess
import tempfile
from ulauncher.api.shared.action.BaseAction import BaseAction
from ulauncher.utils.decorator.run_async import run_async

logger = logging.getLogger(__name__)


class RunScriptAction(BaseAction):
    """
    Runs a user script

    :param str script: script content
    :param list args: arguments
    """

    def __init__(self, script, args=None):
        self.script = script
        self.args = args

    def run(self):
        file = tempfile.NamedTemporaryFile(prefix='ulauncher_RunScript_', delete=False)

        try:
            file.write(self.script.encode())
        except Exception:
            file.close()
            raise
        else:
            file.close()

        try:
            os.chmod(file.name, 0o777)
            logger.debug('Running a script from %s', file.name)
            subprocess.Popen(["%s %s" % (file.name, self.args)],
                             shell=True, stdin=None, stdout=None, stderr=None, close_fds=True)
            self.remove_temp_file(file.name)
        except Exception:
            self.remove_temp_file(file.name)
            raise

    @run_async(daemon=True)
    def remove_temp_file(self, filename: str) -> None:
        time.sleep(1)  # wait just a bit, because Popen runs file asynchronously
        logger.debug('Deleting a temporary file %s', filename)
        os.remove(filename)



from ulauncher.api.shared.action.RenderResultListAction import RenderResultListAction
from ulauncher.search.BaseSearchMode import BaseSearchMode
from ulauncher.search.shortcuts.ShortcutsDb import ShortcutsDb
from ulauncher.search.shortcuts.ShortcutResultItem import ShortcutResultItem


class ShortcutSearchMode(BaseSearchMode):

    def __init__(self):
        self.shortcutsDb = ShortcutsDb.get_instance()

    def is_enabled(self, query):
        """
        Return True if mode should be enabled for a query
        """
        return bool(self._get_active_shortcut(query))

    def _get_active_shortcut(self, query):
        for s in self.shortcutsDb.get_shortcuts():
            if query == s.get('keyword') and s.get('run_without_argument'):
                return s
            if query.startswith('%s ' % s.get('keyword')):
                return s

        return None

    def _create_items(self, shortcuts, default_search=False):
        return [ShortcutResultItem(default_search=default_search, **s) for s in shortcuts]

    def handle_query(self, query):
        """
        @return Action object
        """
        shortcut = self._get_active_shortcut(query)
        if not shortcut:
            raise Exception('No active shortcut. This line should not be entered')

        return RenderResultListAction([ShortcutResultItem(**shortcut)])

    def get_default_items(self):
        return self._create_items([s for s in self.shortcutsDb.get_shortcuts() if s['is_default_search']],
                                  default_search=True)

    def get_searchable_items(self):
        return self._create_items(self.shortcutsDb.get_shortcuts())
